/* TroiaPreview
 * version 0.1 date 2004-09-01
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 *
 * Parameters:
 *   "Log":          generate Logfile
 *   "Res640x480":   use display-resolution 640 x 480
 *   "Res800x600":   use display-resolution 800 x 600
 *   "Res1024x768":  use display-resolution 1024 x 768
 *   "NoFullScreen": do not use fullscreen-mode
 *   "NoMouse":      do not use mouse
 *   "InvertMouse":  invert vertical mouse-axis
 */

#include <windows.h>
#include <stdio.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <math.h>
#include "jpeglib.h"

#define Pi 3.1415926535897932384626433832795
#define PixelX 104
#define PixelY 80
#define PixelZ 32
#define PixelSize 0.14
#define PosMinX PixelSize
#define PosMaxX ((PixelX - 1) * PixelSize)
#define PosMinY PixelSize
#define PosMaxY ((PixelY - 1) * PixelSize)
#define PosMinZ PixelSize
#define PosMaxZ ((PixelZ - 1) * PixelSize)
#define FrontAngleMax (0.2 * Pi)
#define CenterX (0.5 * PixelX * PixelSize)
#define CenterY (0.5 * PixelY * PixelSize)
#define CenterZ (0.5 * PixelZ * PixelSize) 
#define CamptSpeedRange 10.0
#define MoveStep 0.14
#define TurnStep (Pi / 40.0)
#define MouseMoveStep 0.014
#define MouseTurnStep (Pi / 800.0)
#define BackTexSizeX 256 //size of a background texture
#define BackTexSizeY 256
#define BackTexCntX 4 //number of background textures
#define BackTexCntY 3
#define BackTexCnt (BackTexCntX * BackTexCntY)
#define BackTexFirstNo 1 //number of first background-texture
#define ConsoleTexFirstNo (BackTexFirstNo + BackTexCnt) //number of first (and only) console texture
#define ConsoleLeftRightStep 15
#define ConsoleUpDownStep 3
#define ConsolePageStep (ConsoleTextShowSizeY - 2)
#define PixTexSizeX 64 //size of a pixel texture
#define PixTexSizeY 64
#define PixelTexFirstNo (ConsoleTexFirstNo + 1) //number of first (and only) pixel texture
#define TimeRange 86400.0
#define ScriptTimerTime 20
#define ScriptBufferSizeMax 1048576 //maximum size of script-buffer
#define MaxScriptCommandsAtSameTime 20 //maximum number of short command that are executed at the same time

//global types
struct tVector //a vector
{
	double X, Y, Z;
};
struct tCampt //a cam-point
{
	tVector Pos; //position of cam
	tVector Front; //front-direction of cam
};

//global variables
HINSTANCE AppHandle = NULL; //handle of the application
FILE * LogFile = NULL;
bool FullScreen = true, MouseModeOn = true;
HWND WindowHandle; //handle of window
HDC WindowDeviceHandle; //handle of devie-context of window
HGLRC OpenGlContextHandle; //handle of OpenGl-context
long ScreenX; //display-resolution
long ScreenY;
#define ScreenY_X ((float)ScreenY / (float)ScreenX)
tVector Pos, Front; //viewer-position and front-direction
long InvertMouse = 1; //invert vertical mouse-axis? (1 or -1)
bool Back = false; //display background?
bool AntiAlias = false; //use antialiasing?
long CamptsCnt; //list with cam-points
tCampt * pCampts;
double CamptSpeed; //cam-point-speed (in cam-points per second)
double CamptRefTime; //reference-time for cam-point-movement
bool CamMoving; //if cam is moving or not
#define TexCnt 10 //textures
char ScriptMode = 'R'; //S=stopped, W=waiting, R=running
double ScriptWaitTime = 0.0; //time to wait for in ScriptMode is W
char * ScriptBuffer = NULL; //buffer with script-lines to process
long TimeRefTick; //variables for time
double TimeOffset;

//current frame
unsigned char Wall1[PixelX][PixelZ];
unsigned char Wall2[PixelY][PixelZ];
unsigned char Wall3[PixelX][PixelZ];
unsigned char Wall4[PixelY][PixelZ];
unsigned char Ceiling[PixelX][PixelY];
unsigned char Floor[PixelX][PixelY];

//font FixedSys
#define FontSizeX 8
#define FontSizeY 15
unsigned char FontData[256][15] =
{
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 24, 60, 60, 60, 24, 24, 0, 24, 24, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 54, 54, 127, 54, 54, 54, 127, 54, 54, 0, 0, 0 },
	{ 0, 24, 24, 60, 102, 96, 48, 24, 12, 6, 102, 60, 24, 24, 0 },
	{ 0, 0, 112, 216, 218, 118, 12, 24, 48, 110, 91, 27, 14, 0, 0 },
	{ 0, 0, 0, 56, 108, 108, 56, 96, 111, 102, 102, 59, 0, 0, 0 },
	{ 0, 0, 0, 24, 24, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 12, 24, 24, 48, 48, 48, 48, 48, 24, 24, 12, 0 },
	{ 0, 0, 0, 48, 24, 24, 12, 12, 12, 12, 12, 24, 24, 48, 0 },
	{ 0, 0, 0, 0, 0, 54, 28, 127, 28, 54, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 24, 24, 126, 24, 24, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28, 28, 12, 24, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 126, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28, 28, 0, 0, 0 },
	{ 0, 0, 0, 6, 6, 12, 12, 24, 24, 48, 48, 96, 96, 0, 0 },
	{ 0, 0, 0, 30, 51, 55, 55, 51, 59, 59, 51, 30, 0, 0, 0 },
	{ 0, 0, 0, 12, 28, 124, 12, 12, 12, 12, 12, 12, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 6, 12, 24, 48, 96, 126, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 6, 28, 6, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 48, 48, 54, 54, 54, 102, 127, 6, 6, 0, 0, 0 },
	{ 0, 0, 0, 126, 96, 96, 96, 124, 6, 6, 12, 120, 0, 0, 0 },
	{ 0, 0, 0, 28, 24, 48, 124, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 126, 6, 12, 12, 24, 24, 48, 48, 48, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 118, 60, 110, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 102, 102, 62, 12, 24, 56, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 28, 28, 0, 0, 0, 28, 28, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 28, 28, 0, 0, 0, 28, 28, 12, 24, 0 },
	{ 0, 0, 0, 6, 12, 24, 48, 96, 48, 24, 12, 6, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 126, 0, 126, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 96, 48, 24, 12, 6, 12, 24, 48, 96, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 12, 24, 24, 0, 24, 24, 0, 0, 0 },
	{ 0, 0, 0, 126, 195, 195, 207, 219, 219, 207, 192, 127, 0, 0, 0 },
	{ 0, 0, 0, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 124, 102, 102, 102, 124, 102, 102, 102, 124, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 96, 96, 96, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 120, 108, 102, 102, 102, 102, 102, 108, 120, 0, 0, 0 },
	{ 0, 0, 0, 126, 96, 96, 96, 124, 96, 96, 96, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 96, 96, 96, 124, 96, 96, 96, 96, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 96, 96, 110, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 102, 102, 126, 102, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 60, 24, 24, 24, 24, 24, 24, 24, 60, 0, 0, 0 },
	{ 0, 0, 0, 6, 6, 6, 6, 6, 6, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 108, 108, 120, 108, 108, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 96, 96, 96, 96, 96, 96, 96, 96, 126, 0, 0, 0 },
	{ 0, 0, 0, 99, 99, 119, 107, 107, 107, 99, 99, 99, 0, 0, 0 },
	{ 0, 0, 0, 99, 99, 115, 123, 111, 103, 99, 99, 99, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 124, 102, 102, 102, 124, 96, 96, 96, 96, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 12, 6, 0 },
	{ 0, 0, 0, 124, 102, 102, 102, 124, 108, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 96, 48, 24, 12, 6, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 126, 24, 24, 24, 24, 24, 24, 24, 24, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 102, 102, 102, 102, 102, 60, 24, 0, 0, 0 },
	{ 0, 0, 0, 99, 99, 99, 107, 107, 107, 54, 54, 54, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 52, 24, 24, 44, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 102, 102, 60, 24, 24, 24, 24, 0, 0, 0 },
	{ 0, 0, 0, 126, 6, 6, 12, 24, 48, 96, 96, 126, 0, 0, 0 },
	{ 0, 0, 0, 60, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 60 },
	{ 0, 0, 0, 96, 96, 48, 48, 24, 24, 12, 12, 6, 6, 0, 0 },
	{ 0, 0, 0, 60, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 60 },
	{ 0, 24, 60, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255 },
	{ 0, 56, 24, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 0, 96, 96, 124, 102, 102, 102, 102, 102, 124, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 60, 102, 96, 96, 96, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 6, 6, 62, 102, 102, 102, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 60, 102, 102, 126, 96, 96, 60, 0, 0, 0 },
	{ 0, 0, 0, 30, 48, 48, 48, 126, 48, 48, 48, 48, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 62, 102, 102, 102, 102, 102, 62, 6, 6, 124 },
	{ 0, 0, 0, 96, 96, 124, 102, 102, 102, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 24, 24, 0, 120, 24, 24, 24, 24, 24, 126, 0, 0, 0 },
	{ 0, 0, 12, 12, 0, 60, 12, 12, 12, 12, 12, 12, 12, 12, 120 },
	{ 0, 0, 0, 96, 96, 102, 102, 108, 120, 108, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 120, 24, 24, 24, 24, 24, 24, 24, 126, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 126, 107, 107, 107, 107, 107, 99, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 124, 102, 102, 102, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 60, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 124, 102, 102, 102, 102, 102, 124, 96, 96, 96 },
	{ 0, 0, 0, 0, 0, 62, 102, 102, 102, 102, 102, 62, 6, 6, 6 },
	{ 0, 0, 0, 0, 0, 102, 110, 112, 96, 96, 96, 96, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 62, 96, 96, 60, 6, 6, 124, 0, 0, 0 },
	{ 0, 0, 0, 48, 48, 126, 48, 48, 48, 48, 48, 30, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 102, 102, 102, 102, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 102, 102, 102, 102, 102, 60, 24, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 99, 107, 107, 107, 107, 54, 54, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 102, 102, 60, 24, 60, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 102, 102, 102, 102, 102, 102, 60, 12, 24, 240 },
	{ 0, 0, 0, 0, 0, 126, 6, 12, 24, 48, 96, 126, 0, 0, 0 },
	{ 0, 0, 0, 12, 24, 24, 24, 48, 96, 48, 24, 24, 24, 12, 0 },
	{ 0, 0, 0, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24 },
	{ 0, 0, 0, 48, 24, 24, 24, 12, 6, 12, 24, 24, 24, 48, 0 },
	{ 0, 0, 0, 113, 219, 142, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 12, 24, 28, 28, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 28, 28, 12, 24, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 126, 126, 126, 126, 126, 126, 126, 126, 126, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 24, 24, 0, 24, 24, 60, 60, 60, 24 },
	{ 0, 0, 0, 24, 24, 60, 102, 96, 96, 102, 60, 24, 24, 0, 0 },
	{ 0, 0, 0, 60, 102, 96, 96, 252, 96, 96, 192, 254, 0, 0, 0 },
	{ 0, 0, 0, 0, 102, 60, 102, 102, 60, 102, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 102, 102, 102, 60, 126, 24, 126, 24, 24, 0, 0, 0 },
	{ 0, 0, 0, 24, 24, 24, 24, 24, 0, 0, 24, 24, 24, 24, 24 },
	{ 0, 0, 0, 60, 102, 96, 56, 60, 102, 102, 60, 28, 6, 102, 60 },
	{ 0, 0, 102, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 195, 153, 165, 161, 165, 153, 195, 126, 0, 0, 0 },
	{ 0, 0, 0, 60, 6, 62, 102, 62, 0, 126, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 51, 102, 204, 102, 51, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 126, 6, 6, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 126, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 126, 195, 185, 165, 165, 185, 165, 195, 126, 0, 0, 0 },
	{ 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 60, 102, 102, 60, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 24, 24, 126, 24, 24, 0, 126, 0, 0, 0 },
	{ 0, 0, 0, 56, 12, 24, 48, 60, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 56, 12, 24, 12, 56, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 28, 24, 48, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 102, 102, 102, 102, 102, 102, 123, 96, 96, 192 },
	{ 0, 0, 0, 31, 62, 126, 126, 126, 62, 30, 6, 6, 6, 6, 6 },
	{ 0, 0, 0, 0, 0, 0, 28, 28, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24, 12, 56 },
	{ 0, 0, 0, 24, 56, 24, 24, 24, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 102, 60, 0, 126, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 204, 102, 51, 102, 204, 0, 0, 0 },
	{ 0, 0, 96, 224, 99, 102, 108, 24, 55, 111, 219, 31, 3, 0, 0 },
	{ 0, 0, 96, 224, 99, 102, 108, 24, 62, 99, 198, 12, 15, 0, 0 },
	{ 0, 0, 224, 48, 99, 54, 236, 24, 55, 111, 219, 31, 3, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 24, 24, 0, 24, 24, 48, 102, 102, 60 },
	{ 48, 24, 0, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 12, 24, 0, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 60, 102, 0, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 59, 110, 0, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 102, 102, 0, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 60, 102, 60, 24, 60, 102, 102, 102, 126, 102, 102, 102, 0, 0, 0 },
	{ 0, 0, 0, 15, 28, 60, 108, 111, 124, 108, 108, 111, 0, 0, 0 },
	{ 0, 0, 0, 60, 102, 102, 96, 96, 96, 102, 102, 60, 24, 12, 56 },
	{ 48, 24, 0, 126, 96, 96, 96, 124, 96, 96, 96, 126, 0, 0, 0 },
	{ 12, 24, 0, 126, 96, 96, 96, 124, 96, 96, 96, 126, 0, 0, 0 },
	{ 60, 102, 0, 126, 96, 96, 96, 124, 96, 96, 96, 126, 0, 0, 0 },
	{ 102, 102, 0, 126, 96, 96, 96, 124, 96, 96, 96, 126, 0, 0, 0 },
	{ 48, 24, 0, 60, 24, 24, 24, 24, 24, 24, 24, 60, 0, 0, 0 },
	{ 12, 24, 0, 60, 24, 24, 24, 24, 24, 24, 24, 60, 0, 0, 0 },
	{ 60, 102, 0, 60, 24, 24, 24, 24, 24, 24, 24, 60, 0, 0, 0 },
	{ 102, 102, 0, 60, 24, 24, 24, 24, 24, 24, 24, 60, 0, 0, 0 },
	{ 0, 0, 0, 120, 108, 102, 102, 246, 102, 102, 108, 120, 0, 0, 0 },
	{ 59, 110, 0, 99, 99, 115, 123, 111, 103, 99, 99, 99, 0, 0, 0 },
	{ 48, 24, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 12, 24, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 60, 102, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 59, 110, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 102, 102, 0, 60, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 0, 66, 102, 60, 24, 60, 102, 66, 0, 0, 0, 0 },
	{ 0, 0, 0, 62, 102, 110, 110, 126, 118, 118, 102, 124, 0, 0, 0 },
	{ 48, 24, 0, 102, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 12, 24, 0, 102, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 60, 102, 0, 102, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 102, 102, 0, 102, 102, 102, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 12, 24, 0, 102, 102, 102, 102, 60, 24, 24, 24, 24, 0, 0, 0 },
	{ 0, 0, 0, 96, 96, 124, 102, 102, 102, 124, 96, 96, 0, 0, 0 },
	{ 0, 0, 0, 56, 108, 108, 108, 102, 102, 102, 102, 108, 0, 0, 0 },
	{ 0, 112, 48, 24, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 14, 12, 24, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 24, 60, 102, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 113, 219, 142, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 102, 102, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 60, 102, 60, 0, 60, 6, 6, 62, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 110, 27, 27, 127, 216, 216, 119, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 60, 102, 96, 96, 96, 102, 60, 24, 12, 56 },
	{ 0, 112, 48, 24, 0, 60, 102, 102, 126, 96, 96, 60, 0, 0, 0 },
	{ 0, 14, 12, 24, 0, 60, 102, 102, 126, 96, 96, 60, 0, 0, 0 },
	{ 0, 24, 60, 102, 0, 60, 102, 102, 126, 96, 96, 60, 0, 0, 0 },
	{ 0, 0, 102, 102, 0, 60, 102, 102, 126, 96, 96, 60, 0, 0, 0 },
	{ 0, 112, 48, 24, 0, 120, 24, 24, 24, 24, 24, 126, 0, 0, 0 },
	{ 0, 14, 12, 24, 0, 120, 24, 24, 24, 24, 24, 126, 0, 0, 0 },
	{ 0, 24, 60, 102, 0, 120, 24, 24, 24, 24, 24, 126, 0, 0, 0 },
	{ 0, 0, 102, 102, 0, 120, 24, 24, 24, 24, 24, 126, 0, 0, 0 },
	{ 0, 0, 0, 54, 24, 108, 30, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 113, 219, 142, 0, 124, 102, 102, 102, 102, 102, 102, 0, 0, 0 },
	{ 0, 112, 48, 24, 0, 60, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 14, 12, 24, 0, 60, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 24, 60, 102, 0, 60, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 113, 219, 142, 0, 60, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 102, 102, 0, 60, 102, 102, 102, 102, 102, 60, 0, 0, 0 },
	{ 0, 0, 0, 0, 24, 24, 0, 126, 0, 24, 24, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 62, 102, 110, 126, 118, 102, 124, 0, 0, 0 },
	{ 0, 112, 48, 24, 0, 102, 102, 102, 102, 102, 102, 62, 0, 0, 0 },
	{ 0, 14, 12, 24, 0, 102, 102, 102, 102, 102, 102, 62, 0, 0, 0 },
	{ 0, 24, 60, 102, 0, 102, 102, 102, 102, 102, 102, 62, 0, 0, 0 },
	{ 0, 0, 102, 102, 0, 102, 102, 102, 102, 102, 102, 62, 0, 0, 0 },
	{ 0, 14, 12, 24, 0, 102, 102, 102, 102, 102, 102, 60, 12, 24, 240 },
	{ 0, 0, 0, 96, 96, 96, 124, 102, 102, 102, 102, 124, 96, 96, 96 },
	{ 0, 0, 102, 102, 0, 102, 102, 102, 102, 102, 102, 60, 12, 24, 240 }
};
unsigned char CursorData[15] =
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255 };

//console
bool Console = true, ConsoleChanged = true;
#define ConsoleTexSizeX 256 //size of the console-texture
#define ConsoleTexSizeY 256
#define ConsoleColorBorder 2
#define ConsoleSpaceBorder 4
#define ConsoleBorder (ConsoleColorBorder + ConsoleSpaceBorder)
#define ConsoleInnerSizeX (ConsoleTexSizeX - 2 * ConsoleBorder)
#define ConsoleInnerSizeY (ConsoleTexSizeY - 2 * ConsoleBorder)
#define ConsoleTexLineSize ((ConsoleTexSizeX * 3 + 3) & 0xFFFFFFFC)
unsigned char ConsoleData[ConsoleTexSizeY * ConsoleTexLineSize];
#define ConsolePixelComponent( Y, X, I ) ConsoleData[(Y) * ConsoleTexLineSize + (X) * 3 + (I)]
#define ConsoleBackR 0x00
#define ConsoleBackG 0x00
#define ConsoleBackB 0x00
#define ConsoleBorderR 0xC0
#define ConsoleBorderG 0xC0
#define ConsoleBorderB 0xC0
#define ConsoleTextR 0xFF
#define ConsoleTextG 0xFF
#define ConsoleTextB 0x00
#define ConsoleCursorR 0xFF
#define ConsoleCursorG 0x00
#define ConsoleCursorB 0xFF
#define ConsoleTextShowSizeX (long)(ConsoleInnerSizeX / FontSizeX)
#define ConsoleTextShowSizeY (long)(ConsoleInnerSizeY / FontSizeY)
char ConsoleTextShow[ConsoleTextShowSizeY][ConsoleTextShowSizeX];
long ConsoleCursorShowX, ConsoleCursorShowY;
long ConsoleTopLeftX, ConsoleTopLeftY;
#define ConsoleTextSizeX 256
#define ConsoleTextSizeY 256
char ConsoleText[ConsoleTextSizeY][ConsoleTextSizeX];
long ConsoleCursorX, ConsoleCursorY;

//stristr
static char * stristr( char * Str, char * Search )
{
	long I, L;
	L = strlen( Search );
	for( I = 0; Str[I] != 0; I++ )
		if( strnicmp( Str + I, Search, L ) == 0 )
			return Str + I;
	return NULL;
}

//absolute value
static double Abs( double X )
{
	if( X < 0.0 )
		return -X;
	return X;
}

//sign of value
static double Sign( double X )
{
	if( X > 0.0 )
		return 1.0;
	if( X < 0.0 )
		return -1.0;
	return 0.0;
}

//check range of a char
static char CharMinMax( char Value, char Min, char Max )
{
	if( Value < Min ) Value = Min;
	if( Value > Max ) Value = Max;
	return Value;
}

//check range of a short
static short ShortMinMax( short Value, short Min, short Max )
{
	if( Value < Min ) Value = Min;
	if( Value > Max ) Value = Max;
	return Value;
}

//check range of a long
static long LongMinMax( long Value, long Min, long Max )
{
	if( Value < Min ) Value = Min;
	if( Value > Max ) Value = Max;
	return Value;
}

//check range of a float
static float FloatMinMax( float Value, float Min, float Max )
{
	if( Value < Min ) Value = Min;
	if( Value > Max ) Value = Max;
	return Value;
}

//check range of a double
static double DoubleMinMax( double Value, double Min, double Max )
{
	if( Value < Min ) Value = Min;
	if( Value > Max ) Value = Max;
	return Value;
}

//set vector
static void SetVector( tVector * pDest, double X, double Y, double Z )
{
	pDest->X = X;
	pDest->Y = Y;
	pDest->Z = Z;
}

//vector assginment
static void Copy( tVector * pDest, tVector Src )
{
	pDest->X = Src.X;
	pDest->Y = Src.Y;
	pDest->Z = Src.Z;
}

//vector compare
static bool Equal( tVector V1, tVector V2 )
{
	if( V1.X != V2.X )
		return false;
	if( V1.Y != V2.Y )
		return false;
	if( V1.Z != V2.Z )
		return false;
	return true;
}

//vector addition
static void Add( tVector * pDest, tVector Src1, tVector Src2 )
{
	pDest->X = Src1.X + Src2.X;
	pDest->Y = Src1.Y + Src2.Y;
	pDest->Z = Src1.Z + Src2.Z;
}

//vector subtraction
static void Sub( tVector * pDest, tVector Src1, tVector Src2 )
{
	pDest->X = Src1.X - Src2.X;
	pDest->Y = Src1.Y - Src2.Y;
	pDest->Z = Src1.Z - Src2.Z;
}

//scale a vector
static void Scale( tVector * pV, double Factor )
{
	pV->X *= Factor;
	pV->Y *= Factor;
	pV->Z *= Factor;
}

//linear-combine two vectors
static void LinearCombine( tVector * pDest, double Factor1, tVector Src1, double Factor2, tVector Src2 )
{
	pDest->X = Factor1 * Src1.X + Factor2 * Src2.X;
	pDest->Y = Factor1 * Src1.Y + Factor2 * Src2.Y;
	pDest->Z = Factor1 * Src1.Z + Factor2 * Src2.Z;
}

//length of a vector
static double Len( tVector V )
{
	return sqrt( V.X * V.X + V.Y * V.Y + V.Z * V.Z );
}

//normalize a vector
static void Normalize( tVector * pV )
{
	double L;
	L = Len( *pV );
	if( L == 0.0 )
		pV->X = 0.0, pV->Y = 0.0, pV->Z = 1.0;
	else
		Scale( pV, 1.0 / L );
}

//scalar product
static double ScalarProduct( tVector Vec1, tVector Vec2 )
{
	return Vec1.X * Vec2.X + Vec1.Y * Vec2.Y + Vec1.Z * Vec2.Z;
}

//angle between two vectors
static double VectorAngle( tVector Vec1, tVector Vec2 )
{
	return acos( ScalarProduct( Vec1, Vec2 ) / ScalarProduct( Vec2, Vec2 ) );
}

//vector product
static void VectorProduct( tVector * pDest, tVector Vec1, tVector Vec2 )
{
	pDest->X = Vec1.Y * Vec2.Z - Vec1.Z * Vec2.Y;
	pDest->Y = Vec1.Z * Vec2.X - Vec1.X * Vec2.Z;
	pDest->Z = Vec1.X * Vec2.Y - Vec1.Y * Vec2.X;
}

//rotate vector V around vector A
static void Rotate( tVector * pV, tVector A, double Angle )
{
	//local variables
	tVector V;
	double R;

	//normalize axis A
	Normalize( &A );

	//turn around X-Axis
	Copy( &V, *pV );
	R = Angle * A.X;
	pV->Y = cos( R ) * V.Y + sin( R ) * V.Z;
	pV->Z = -sin( R ) * V.Y + cos( R ) * V.Z;

	//turn around Y-Axis
	Copy( &V, *pV );
	R = Angle * A.Y;
	pV->X = cos( R ) * V.X - sin( R ) * V.Z;
	pV->Z = sin( R ) * V.X + cos( R ) * V.Z;

	//turn around Z-Axis
	Copy( &V, *pV );
	R = Angle * A.Z;
	pV->X = cos( R ) * V.X + sin( R ) * V.Y;
	pV->Y = -sin( R ) * V.X + cos( R ) * V.Y;
}

//set time
static void TimeSet( double Time )
{
	TimeRefTick = GetTickCount( );
	TimeOffset = Time;
}

//get time
static double TimeGet( )
{
	long Tick;
	Tick = GetTickCount( ) - TimeRefTick;
	return (double)Tick / 1000.0 + TimeOffset;
}

//write a line to the logfile
static void Log( char * Info, char * Text )
{
	//local variables
	SYSTEMTIME SysTime;
	char Timestamp[64];

	//logfile not open, no info or no text
	if( LogFile == NULL || Info[0] == 0 || Text[0] == 0 )
		return;

	//write timestamp
	GetLocalTime( &SysTime );
	sprintf( Timestamp, "%04d-%02d-%02d %02d:%02d:%02d ",
		SysTime.wYear, SysTime.wMonth, SysTime.wDay, SysTime.wHour, SysTime.wMinute, SysTime.wSecond );
	fwrite( Timestamp, sizeof( char ), strlen( Timestamp ), LogFile );

	//write Info
	fwrite( Info, sizeof( char ), strlen( Info ), LogFile );

	//write seperator
	fwrite( ": ", sizeof( char ), 2, LogFile );

	//write Text
	fwrite( Text, sizeof( char ), strlen( Text ), LogFile );

	//write newline
	fwrite( "\n", sizeof( char ), 1, LogFile );
}

//show a pixel
//Pos: top left corner of pixel
//Right: vector from left to right side
//Bottom: vector from top to bottom side
//Color: greyscale value of pixel (0..255)
static void ShowPixel( tVector Pos, tVector Right, tVector Bottom, unsigned char Color )
{
	float fColor;

	//set Color
	fColor = (float)Color * (1.0f / 255.0f);
	glColor3f( fColor, fColor, fColor );

	//draw a quad
	glBegin( GL_QUADS );
	glTexCoord2d( 0.0, 0.0 ); //top left
	glVertex3d( Pos.X, Pos.Y, Pos.Z );
	Add( &Pos, Pos, Right ); //top right
	glTexCoord2d( 1.0, 0.0 );
	glVertex3d( Pos.X, Pos.Y, Pos.Z );
	Add( &Pos, Pos, Bottom ); //bottom right
	glTexCoord2d( 1.0, 1.0 );
	glVertex3d( Pos.X, Pos.Y, Pos.Z );
	Sub( &Pos, Pos, Right ); //bottom left
	glTexCoord2d( 0.0, 1.0 );
	glVertex3d( Pos.X, Pos.Y, Pos.Z );
	glEnd( );
}

//show pixels
static void ShowPixels( )
{
	//local variables
	long X, Y, Z;
	tVector Pos, Right, Bottom;

	//bind pixel texture
	glBindTexture( GL_TEXTURE_2D, PixelTexFirstNo );
	//enable texture
	glEnable( GL_TEXTURE_2D );
	//set polygon mode
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	//show wall1
	Pos.Y = PixelY * PixelSize;
	SetVector( &Right, PixelSize, 0.0, 0.0 );
	SetVector( &Bottom, 0.0, 0.0, -PixelSize );
	for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
		for( X = 0, Pos.X = 0.0; X < PixelX; X++, Pos.X += PixelSize )
			ShowPixel( Pos, Right, Bottom, Wall1[Z][X] );
	//show wall2
	Pos.X = PixelX * PixelSize;
	SetVector( &Right, 0.0, -PixelSize, 0.0 );
	SetVector( &Bottom, 0.0, 0.0, -PixelSize );
	for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
		for( Y = 0, Pos.Y = PixelY * PixelSize; Y < PixelY; Y++, Pos.Y -= PixelSize )
			ShowPixel( Pos, Right, Bottom, Wall2[Z][Y] );
	//show wall3
	Pos.Y = 0.0;
	SetVector( &Right, -PixelSize, 0.0, 0.0 );
	SetVector( &Bottom, 0.0, 0.0, -PixelSize );
	for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
		for( X = 0, Pos.X = PixelX * PixelSize; X < PixelX; X++, Pos.X -= PixelSize )
			ShowPixel( Pos, Right, Bottom, Wall3[Z][X] );
	//show wall4
	Pos.X = 0.0;
	SetVector( &Right, 0.0, PixelSize, 0.0 );
	SetVector( &Bottom, 0.0, 0.0, -PixelSize );
	for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
		for( Y = 0, Pos.Y = 0.0; Y < PixelY; Y++, Pos.Y += PixelSize )
			ShowPixel( Pos, Right, Bottom, Wall4[Z][Y] );
	//show ceiling
	Pos.Z = PixelZ * PixelSize;
	SetVector( &Right, PixelSize, 0.0, 0.0 );
	SetVector( &Bottom, 0.0, PixelSize, 0.0 );
	for( Y = 0, Pos.Y = (PixelY - 1) * PixelSize; Y < PixelY; Y++, Pos.Y -= PixelSize )
		for( X = 0, Pos.X = 0.0; X < PixelX; X++, Pos.X += PixelSize )
			ShowPixel( Pos, Right, Bottom, Ceiling[Y][X] );
	//show floor
	Pos.Z = 0.0;
	SetVector( &Right, PixelSize, 0.0, 0.0 );
	SetVector( &Bottom, 0.0, -PixelSize, 0.0 );
	for( Y = 0, Pos.Y = PixelY * PixelSize; Y < PixelY; Y++, Pos.Y -= PixelSize )
		for( X = 0, Pos.X = 0.0; X < PixelX; X++, Pos.X += PixelSize )
			ShowPixel( Pos, Right, Bottom, Floor[Y][X] );
}

//show picture
static void Show( )
{
	//local variables
	long X, Y, I;
	float fX, fY, fDX, fDY;

	//clear picture and depth-information
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	//settings for background and console
	glMatrixMode( GL_PROJECTION ); //set projection
	glLoadIdentity( );
	glMatrixMode( GL_MODELVIEW ); //set view
	glLoadIdentity( );
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL ); //draw filled polygons
	glEnable( GL_TEXTURE_2D ); //textures on

	//no antialiasing
	glDisable( GL_POINT_SMOOTH );
	glDisable( GL_LINE_SMOOTH );
	glDisable( GL_POLYGON_SMOOTH );
	glDisable( GL_BLEND );

	//draw background
	if( Back )
	{
		glColor3f( 1.0f, 1.0f, 1.0f );
		fDX = 2.0f / (float)BackTexCntX;
		fDY = 2.0f / (float)BackTexCntY;
		for( Y = 0; Y < BackTexCntY; Y++ )
		for( X = 0; X < BackTexCntX; X++ )
		{
			I = Y * BackTexCntX + X;
			fX = -1.0f + X * fDX;
			fY = 1.0f - Y * fDY;
			glBindTexture( GL_TEXTURE_2D, BackTexFirstNo + I );
			glBegin( GL_QUADS );
			glNormal3f( 0.0f, 0.0f, 1.0f );
			glTexCoord2f( 0.0f, 0.0f );
			glVertex3f( fX, fY, 1.0f );
			glTexCoord2f( 1.0f, 0.0f );
			glVertex3f( fX + fDX, fY, 1.0f );
			glTexCoord2f( 1.0f, 1.0f );
			glVertex3f( fX + fDX, fY - fDY, 1.0f );
			glTexCoord2f( 0.0f, 1.0f );
			glVertex3f( fX, fY - fDY, 1.0f );
			glEnd( );
		}
	}

	//draw console
	if( Console )
	{
		fX = 512.0f / (float)ScreenX;
		fY = 512.0f / (float)ScreenY;
		glColor3f( 1.0f, 1.0f, 1.0f );
		glBindTexture( GL_TEXTURE_2D, ConsoleTexFirstNo );
		glBegin( GL_QUADS );
		glNormal3f( 0.0f, 0.0f, 1.0f );
		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( -1.0f, -1.0f, -1.0f );
		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( -1.0f + fX, -1.0f, -1.0f );
		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( -1.0f + fX, -1.0f + fY, -1.0f );
		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( -1.0f, -1.0f + fY, -1.0f );
		glEnd( );
	}

	//settings for objects in space
	glMatrixMode( GL_PROJECTION ); //set projection
	glLoadIdentity( );
	glFrustum( -0.01, 0.01, -0.01 * ScreenY_X, 0.01 * ScreenY_X, 0.01, 20.0 );
	glMatrixMode( GL_MODELVIEW ); //set view
	glLoadIdentity( );
	gluLookAt( Pos.X, Pos.Y, Pos.Z, Pos.X + Front.X, Pos.Y + Front.Y, Pos.Z + Front.Z, 0.0, 0.0, 1.0 );

	//enable antialiasing ?
	if( AntiAlias )
	{
		glEnable( GL_POINT_SMOOTH );
		glEnable( GL_LINE_SMOOTH );
		glEnable( GL_POLYGON_SMOOTH );
		glEnable( GL_BLEND );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		//glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST );
   	}

	//show pixels
	ShowPixels( );

	//show picture
	SwapBuffers( WindowDeviceHandle );
}

//formula for interpolation
//A=f(-1), B=f(0), C=f(1), D=f(2), f(X) wanted
#define Move_f(A,B,C,D) (-A/6.0 + B/2.0 - C/2.0 + D/6.0) * XXX + (A/2.0 - B + C/2.0) * XX + (-A/3.0 - B/2 + C - D/6.0)*X + B

//repair current position and front direction
static void RepairPosFront( )
{
	tVector Dir, Top, Left;
	double Angle;

	//do not escape allowed region
	Pos.X = DoubleMinMax( Pos.X, PosMinX, PosMaxX );
	Pos.Y = DoubleMinMax( Pos.Y, PosMinY, PosMaxY );
	Pos.Z = DoubleMinMax( Pos.Z, PosMinZ, PosMaxZ );

	//normalize front vector
	Normalize( &Front );

	//get direction over ground
	Dir = Front;
	Dir.Z = 0;
	Normalize( &Dir );
	//get top and left vectors
	SetVector( &Top, 0.0, 0.0, 1.0 );
	Left = Dir;
	Rotate( &Left, Top, -0.5 * Pi );

	//check angle between front vector and direction
	Angle = VectorAngle( Front, Dir );
	if( Angle > FrontAngleMax ) //angle too big
	{
		if( Front.Z > 0.0 ) //looking up
		{
			//set front vector to maximum up direction
			Front = Dir;
			Rotate( &Front, Left, FrontAngleMax );
		}
		else //looking down
		{
			//set front vector to maximum down direction
			Front = Dir;
			Rotate( &Front, Left, - FrontAngleMax );
		}
	}
}

//move cam
static void MoveCam( )
{
	//local variables
	double Time, CurPt, X, XX, XXX;
	long CurPtInt;
	tVector P1, P2, P3, P4;

	//cam is not moving
	if( !CamMoving )
		return;

	//get time
	Time = TimeGet( );

	//one cam-point
	if( CamptsCnt == 1 )
	{
		//move camera to that cam-point
		Copy( &Pos, pCampts[0].Pos );
		Copy( &Front, pCampts[0].Front );
	}
	//more cam-points
	else if( CamptsCnt > 1 )
	{
		//get current cam-point-number
		CurPt = CamptSpeed * (Time - CamptRefTime);
		while( CurPt < 0.0 )
			CurPt += CamptsCnt;
		while( CurPt >= CamptsCnt )
			CurPt -= CamptsCnt;
		//get points for interpolation
		CurPtInt = (long)CurPt;
		X = CurPt - (double)CurPtInt;
		XX = X * X;
		XXX = XX * X;
		//get points for pos-interpolation
		Copy( &P1, pCampts[(CurPtInt + CamptsCnt - 1) % CamptsCnt].Pos );
		Copy( &P2, pCampts[CurPtInt].Pos );
		Copy( &P3, pCampts[(CurPtInt + 1) % CamptsCnt].Pos );
		Copy( &P4, pCampts[(CurPtInt + 2) % CamptsCnt].Pos );
		//pos-interpolation
		//P1=f(-1), P2=f(0), P3=f(1), P4=f(2), f(X) wanted
		Pos.X = Move_f( P1.X, P2.X, P3.X, P4.X );
		Pos.Y = Move_f( P1.Y, P2.Y, P3.Y, P4.Y );
		Pos.Z = Move_f( P1.Z, P2.Z, P3.Z, P4.Z );
		//get points for front-interpolation
		Copy( &P1, pCampts[(CurPtInt + CamptsCnt - 1) % CamptsCnt].Front );
		Copy( &P2, pCampts[CurPtInt].Front );
		Copy( &P3, pCampts[(CurPtInt + 1) % CamptsCnt].Front );
		Copy( &P4, pCampts[(CurPtInt + 2) % CamptsCnt].Front );
		//front-interpolation
		//P1=f(-1), P2=f(0), P3=f(1), P4=f(2), f(X) wanted
		Front.X = Move_f( P1.X, P2.X, P3.X, P4.X );
		Front.Y = Move_f( P1.Y, P2.Y, P3.Y, P4.Y );
		Front.Z = Move_f( P1.Z, P2.Z, P3.Z, P4.Z );
	}

	//repair position and front direction
	RepairPosFront( );
}

//load an image from a bitmap-file
static bool LoadBitmapFile( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData )
{
	//local variables
	FILE * File;
	long I, X, Y, Pos, DummyCnt;
	BITMAPFILEHEADER BitmapFileHeader;
	BITMAPINFOHEADER BitmapInfoHeader;
	RGBQUAD * pColors;
	bool BottomUp;
	unsigned char * pData;
	unsigned char CI;
	
	//initialize return-values
	*pSizeX = -1;
	*pSizeY = -1;
	*ppData = NULL;

	//empty filename ===> fopen will raise assertion, so exit
	if( FileName == NULL || FileName[0] == 0 )
		return false;

	//open file
	File = fopen( FileName, "rb" );
	if( File == NULL )
		return false;

	//read bitmap-file-header
	I = fread( &BitmapFileHeader, sizeof( BITMAPFILEHEADER ), 1, File );
	if( I != 1 )
	{
		//close file
		fclose( File );
		return false;
	}
	//check bitmap-file-header
	if( ((char *)&(BitmapFileHeader.bfType))[0] != 'B' || ((char *)&(BitmapFileHeader.bfType))[1] != 'M' )
	{
		//close file
		fclose( File );
		return false;
	}

	//read bitmap-info-header
	I = fread( &BitmapInfoHeader, sizeof( BITMAPINFOHEADER ), 1, File );
	if( I != 1 )
	{
		//close file
		fclose( File );
		return false;
	}
	//check bitmap-info-header
	if( BitmapInfoHeader.biSize != sizeof( BITMAPINFOHEADER ) ||
		BitmapInfoHeader.biPlanes != 1 || //1 plane
		BitmapInfoHeader.biCompression != BI_RGB ) //no compression
	{
		//close file
		fclose( File );
		return false;
	}
	//8 bpp bitmap
	if( BitmapInfoHeader.biBitCount == 8 )
	{
		//check number of used colors
		if( BitmapInfoHeader.biClrUsed == 0 )
			BitmapInfoHeader.biClrUsed = 256;
		if( BitmapInfoHeader.biClrUsed > 256 )
		{
			//close file
			fclose( File );
			return false;
		}
	}
	//24 bpp bitmap
	else if( BitmapInfoHeader.biBitCount == 24 )
	{
		//do not use colors in color-table
		BitmapInfoHeader.biClrUsed = 0;
	}
	//other bitmap-formats than 8 bpp or 24 bpp not supported
	else
	{
		//close file
		fclose( File );
		return false;
	}

	//load color-table
	pColors = (RGBQUAD *)calloc( BitmapInfoHeader.biClrUsed, sizeof( RGBQUAD ) );
	I = fread( pColors, sizeof( RGBQUAD ), BitmapInfoHeader.biClrUsed, File );
	if( I != (long)BitmapInfoHeader.biClrUsed )
	{
		//free color-table
		free( pColors );
		//close file
		fclose( File );
		return false;
	}

	//bottom-up-bitmap
	if( BitmapInfoHeader.biHeight < 0 )
	{
		BitmapInfoHeader.biHeight *= -1;
		BottomUp = false;
	}
	else
	{
		BottomUp = true;
	}

	//go to start of pixels
	if( fseek( File, BitmapFileHeader.bfOffBits, SEEK_SET ) != 0 )
	{
		//free color-table
		free( pColors );
		//close file
		fclose( File );
		return false;
	}

	//allocate buffer for pixels
	pData = (unsigned char *)malloc( BitmapInfoHeader.biHeight * ((BitmapInfoHeader.biWidth * 3 + 3) & 0xFFFFFFFC) );

	//calculate number of dummy-characters between lines (lines are 4-byte-aligned)
	if( BitmapInfoHeader.biBitCount == 8 ) //8 bpp
		DummyCnt = 3 - ((BitmapInfoHeader.biWidth + 3) & 0x00000003); 
	else //24 bpp
		DummyCnt = 3 - ((BitmapInfoHeader.biWidth * 3 + 3) & 0x00000003);
	
	//read image
	for( Y = 0; Y < BitmapInfoHeader.biHeight; Y++ )
	{
		//read line
		for( X = 0; X < BitmapInfoHeader.biWidth; X++ )
		{
			//get position in buffer
			if( BottomUp )
				Pos = (BitmapInfoHeader.biHeight - Y - 1) * ((BitmapInfoHeader.biWidth * 3 + 3) & 0xFFFFFFFC) + X * 3;
			else
				Pos = Y * ((BitmapInfoHeader.biWidth * 3 + 3) & 0xFFFFFFFC) + X * 3;
			//read pixel 8 bpp
			if( BitmapInfoHeader.biBitCount == 8 )
			{
				//read color-index
				if( fread( &CI, 1, 1, File ) != 1 )
					break;
				//check color
				if( CI >= BitmapInfoHeader.biClrUsed )
					break;
				//get color from color-table
				memcpy( pData + Pos, pColors + CI, 3 );
			}
			//read pixel 24 bpp
			else
			{
				//read color
				if( fread( pData + Pos, 1, 3, File ) != 3 )
					break;
			}
		}
		if( X < BitmapInfoHeader.biWidth ) //error in for-loop (X)
			break;
		//go to next 4-byte-boundary
		if( fseek( File, DummyCnt, SEEK_CUR ) != 0 )
			break;
	}
	if( Y < BitmapInfoHeader.biHeight ) //error in for-loop (Y)
	{
		//free bitmap-data
		free( pData );
		//free color-table
		free( pColors );
		//close file
		fclose( File );
		return false;
	}

	//free color-table
	free( pColors );
	//close file
	fclose( File );

	//success
	*pSizeX = BitmapInfoHeader.biWidth;
	*pSizeY = BitmapInfoHeader.biHeight;
	*ppData = pData;
	return true;
}

//load an image from a jpeg-file
static bool LoadJpegFile( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData )
{
	//local variables
	FILE * File;
	jpeg_error_mgr JpegErr;
	jpeg_decompress_struct JpegDecInfo;
	long LineSize, X, Y;
	unsigned char * pData, * pLine, * pPixel, Color;
	
	//initialize return-values
	*pSizeX = -1;
	*pSizeY = -1;
	*ppData = NULL;

	//empty filename ===> fopen will raise assertion, so exit
	if( FileName == NULL || FileName[0] == 0 )
		return false;

	//open file
	File = fopen( FileName, "rb" );
	if( File == NULL )
		return false;

	//initialize jpeg-error
	JpegDecInfo.err = jpeg_std_error( &JpegErr );
	//create jpeg-decompression-object
	jpeg_create_decompress( &JpegDecInfo );
	//setup file-handle from which to read
	jpeg_stdio_src( &JpegDecInfo, File );

	//get infos from file
	jpeg_read_header( &JpegDecInfo, true );
	//set output-format to 3 bytes per pixel and RGB
	JpegDecInfo.output_components = 3;
	JpegDecInfo.out_color_space = JCS_RGB;
	//start decompressing image
	jpeg_start_decompress( &JpegDecInfo );

	//allocate buffer for pixels
	LineSize = ((JpegDecInfo.output_width * 3 + 3) & 0xFFFFFFFC);
	pData = (unsigned char *)malloc( JpegDecInfo.output_height * LineSize );

	//read all scanlines of image and write them into buffer
	while( JpegDecInfo.output_scanline < JpegDecInfo.output_height )
	{
		pLine = pData + JpegDecInfo.output_scanline * LineSize;
		jpeg_read_scanlines( &JpegDecInfo, &pLine, 1 );
	}

	//finish decompression
	jpeg_finish_decompress( &JpegDecInfo );
	//destroy jpeg-decompression-object
	jpeg_destroy_decompress( &JpegDecInfo );

	//close file
	fclose( File );

	//convert from RGB to BGR
	pLine = pData; //first line
	for( Y = 0; Y < (long)JpegDecInfo.output_height; Y++ )
	{
		pPixel = pLine; //first pixel of line
		for( X = 0; X < (long)JpegDecInfo.output_width; X++ )
		{
			Color = pPixel[0]; //exchange red and blue
			pPixel[0] = pPixel[2];
			pPixel[2] = Color;
			pPixel += 3; //next pixel
		}
		pLine += LineSize; //next line
	}

	//success
	*pSizeX = JpegDecInfo.output_width;
	*pSizeY = JpegDecInfo.output_height;
	*ppData = pData;
	return true;
}

//resize an image (old image stays unaltered)
static void ImageResize( long OldSizeX, long OldSizeY, unsigned char * pOldData, long NewSizeX, long NewSizeY, unsigned char * * ppNewData )
{
	//local variables
	long NewX, NewY, FirstX, LastX, FirstY, LastY, OldX, OldY, Pos;
	double FactorX, FactorY, OldX1, OldX2, OldY1, OldY2;
	double FirstXAntiFactor, LastXAntiFactor, FirstYAntiFactor, LastYAntiFactor;
	double LineFactor, PixelFactor, SumB, SumG, SumR;
	unsigned char * pNewData;

	//allocate buffer for new image
	pNewData = (unsigned char *)malloc( NewSizeY * ((NewSizeX * 3 + 3) & 0xFFFFFFFC) );

	//calculate factors
	FactorX = (double)OldSizeX / (double)NewSizeX;
	FactorY = (double)OldSizeY / (double)NewSizeY;
	
	//loop through new image
	for( NewY = 0; NewY < NewSizeY; NewY++ )
	for( NewX = 0; NewX < NewSizeX; NewX++ )
	{
		//the new pixel is (NewX, NewY)-(NewX+1, NewY+1)
		//calculate the coordinates of the old pixels
		OldX1 = (double)NewX * FactorX;
		OldX2 = (double)(NewX+1) * FactorX;
		OldY1 = (double)NewY * FactorY;
		OldY2 = (double)(NewY+1) * FactorY;
		//calculate first and last coordinates and factors
		FirstX = (long)OldX1;
		FirstXAntiFactor = OldX1 - (double)FirstX;
		LastX = (long)OldX2;
		LastXAntiFactor = 1.0 - (OldX2 - (double)LastX);
		FirstY = (long)OldY1;
		FirstYAntiFactor = OldY1 - (double)FirstY;
		LastY = (long)OldY2;
		LastYAntiFactor = 1.0 - (OldY2 - (double)LastY);
		//get pixel-color-sums
		SumB = 0;
		SumG = 0;
		SumR = 0;
		//loop through rectangle
		for( OldY = FirstY; OldY <= LastY; OldY++ )
		{
			//get line-factor
			LineFactor = 1.0;
			if( OldY == FirstY )
				LineFactor -= FirstYAntiFactor;
			if( OldY == LastY )
				LineFactor -= LastYAntiFactor;
			//loop through line
			for( OldX = FirstX; OldX <= LastX; OldX++ )
			{
				//get pixel factor
				PixelFactor = 1.0;
				if( OldX == FirstX )
					PixelFactor -= FirstXAntiFactor;
				if( OldX == LastX )
					PixelFactor -= LastXAntiFactor;
				PixelFactor *= LineFactor;
				//get color-values
				Pos = OldY * ((OldSizeX * 3 + 3) & 0xFFFFFFFC) + OldX * 3;
				SumB += (double)pOldData[Pos + 0] * PixelFactor;
				SumG += (double)pOldData[Pos + 1] * PixelFactor;
				SumR += (double)pOldData[Pos + 2] * PixelFactor;
			}
		}
		//set color-values
		PixelFactor = (OldX2 - OldX1) * (OldY2 - OldY1);
		Pos = NewY * ((NewSizeX * 3 + 3) & 0xFFFFFFFC) + NewX * 3;
		pNewData[Pos + 0] = (unsigned char)(SumB / PixelFactor);
		pNewData[Pos + 1] = (unsigned char)(SumG / PixelFactor);
		pNewData[Pos + 2] = (unsigned char)(SumR / PixelFactor);
	}
	
	//return new image
	*ppNewData = pNewData;
}

//load an image and resize
static bool LoadImageResize( char * FileName, long ReSizeX, long ReSizeY, unsigned char * * ppRetData,
	bool (* LoadImageFileFunction)( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData ) )
{
	//local variables
	long SizeX, SizeY;
	unsigned char * pData, * pResizeData;

	//load image
	if( !LoadImageFileFunction( FileName, &SizeX, &SizeY, &pData ) )
		//return error
		return false;

	//size is correct
	if( SizeX == ReSizeX && SizeY == ReSizeY )
	{
		//return original image
		*ppRetData = pData;
		return true;
	}

	//resize image
	ImageResize( SizeX, SizeY, pData, ReSizeX, ReSizeY, &pResizeData );

	//free original image
	free( pData );

	//return resized image
	*ppRetData = pResizeData;
	return true;
}
	
//load an image in pieces
//total size: (SizeX * CntX) x (SizeY * CntY)
//size of one piece: SizeX x SizeY
//pPieceData: array of unsigend char * to receive pointers to data of pieces
static bool LoadImagePieces( char * FileName, long SizeX, long SizeY, long CntX, long CntY, unsigned char * pPieceData[],
	bool (* LoadImageFileFunction)( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData ) )
{
	//local variables
	unsigned char * pData;
	long X, Y, YY;

	//load bitmap
	if( !LoadImageResize( FileName, SizeX * CntX, SizeY * CntY, &pData, LoadImageFileFunction ) )
		return false;

	//allocate memory for pieces
	for( Y = 0; Y < CntY; Y++ )
	for( X = 0; X < CntX; X++ )
		pPieceData[Y*CntX+X] = (unsigned char *)malloc( SizeY * ((SizeX * 3 + 3) & 0xFFFFFFFC) );

	//cut data into pieces
	for( Y = 0; Y < CntY; Y++ )
	for( X = 0; X < CntX; X++ )
	{
		for( YY = 0; YY < SizeY; YY++ )
			memcpy( pPieceData[Y*CntX+X] + YY * ((SizeX * 3 + 3) & 0xFFFFFFFC),
				pData + (Y * SizeY + YY) * ((SizeX * CntX * 3 + 3) & 0xFFFFFFFC) + X * SizeX * 3,
				SizeX * 3 );
	}

	//free original bitmap-data
	free( pData );

	//return success
	return true;
}

//load texture TexNo from image-data
static void LoadTexture( unsigned char * pData, long SizeX, long SizeY, long TexNo )
{
	//test pointer
	if( pData == NULL )
		return;

	//load texture into OpenGl
	glBindTexture( GL_TEXTURE_2D, TexNo );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexImage2D( GL_TEXTURE_2D, 0, 3, SizeX, SizeY, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, pData );
}

//unload texture TexNo
static void UnloadTexture( long TexNo )
{
	//local variables
	unsigned char BlackPixel[3] = {0, 0, 0};

	//unload texture from OpenGl by loading a one-pixel black one
	glBindTexture( GL_TEXTURE_2D, TexNo );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexImage2D( GL_TEXTURE_2D, 0, 3, 1, 1, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, &BlackPixel );
}

//inistialize console
static void ConsoleInit( )
{
	//local variables
	long X, Y, FntX, FntY, TexX, TexY;

	//clear console-data
	for( Y = 0; Y < ConsoleTexSizeY; Y++ )
	for( X = 0; X < ConsoleTexSizeX; X++ )
	{
		if( X >= ConsoleColorBorder && X < ConsoleTexSizeX - ConsoleColorBorder &&
			Y >= ConsoleColorBorder && Y < ConsoleTexSizeY - ConsoleColorBorder )
		{
			ConsolePixelComponent( Y, X, 0 ) = ConsoleBackB;
			ConsolePixelComponent( Y, X, 1 ) = ConsoleBackG;
			ConsolePixelComponent( Y, X, 2 ) = ConsoleBackR;
		}
		else
		{
			ConsolePixelComponent( Y, X, 0 ) = ConsoleBorderB;
			ConsolePixelComponent( Y, X, 1 ) = ConsoleBorderG;
			ConsolePixelComponent( Y, X, 2 ) = ConsoleBorderR;
		}
	}

	//clear text-show-buffer
	for( Y = 0; Y < ConsoleTextShowSizeY; Y++ )
	for( X = 0; X < ConsoleTextShowSizeX; X++ )
		ConsoleTextShow[Y][X] = 0;

	//clear text-buffer
	for( Y = 0; Y < ConsoleTextSizeY; Y++ )
	for( X = 0; X < ConsoleTextSizeX; X++ )
		ConsoleText[Y][X] = 0;

	//cursor-coordinates
	ConsoleCursorShowX = 0;
	ConsoleCursorShowY = 0;
	ConsoleTopLeftX = 0;
	ConsoleTopLeftY = 0;
	ConsoleCursorX = 0;
	ConsoleCursorY = 0;

	//put cursor into console-buffer
	for( FntY = 0; FntY < FontSizeY; FntY++ )
	for( FntX = 0; FntX < FontSizeX; FntX++ )
	{
		TexX = ConsoleCursorX * FontSizeX + FntX + ConsoleBorder;
		TexY = ConsoleInnerSizeY - (ConsoleCursorY * FontSizeY + FntY) + ConsoleBorder;
		if( (CursorData[FntY] & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
		{
			ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleCursorB;
			ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleCursorG;
			ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleCursorR;
		}
		else
		{
			ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
			ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
			ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
		}
	}

	//console has changed
	ConsoleChanged = true;
}

//check (and correct) position in console
static void ConsoleCheckPos( )
{
	//correct cursor-position
	if( ConsoleCursorX < 0 )
		ConsoleCursorX = 0;
	if( ConsoleCursorX > ConsoleTextSizeX - 1 )
		ConsoleCursorX = ConsoleTextSizeX - 1;
	if( ConsoleCursorY < 0 )
		ConsoleCursorY = 0;
	if( ConsoleCursorY > ConsoleTextSizeY - 1 )
		ConsoleCursorY = ConsoleTextSizeY - 1;

	//correct top-left-corner-position
	if( ConsoleTopLeftX < 0 )
		ConsoleTopLeftX = 0;
	if( ConsoleTopLeftX > ConsoleTextSizeX - ConsoleTextShowSizeX )
		ConsoleTopLeftX = ConsoleTextSizeX - ConsoleTextShowSizeX;
	if( ConsoleTopLeftY < 0 )
		ConsoleTopLeftY = 0;
	if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
		ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
}

//check (and correct) position in console and adjust so that cursor is visible
static void ConsoleCheckPosCursorVisible( )
{
	//correct cursor-position
	if( ConsoleCursorX < 0 )
		ConsoleCursorX = 0;
	if( ConsoleCursorX > ConsoleTextSizeX - 1 )
		ConsoleCursorX = ConsoleTextSizeX - 1;
	if( ConsoleCursorY < 0 )
		ConsoleCursorY = 0;
	if( ConsoleCursorY > ConsoleTextSizeY - 1 )
		ConsoleCursorY = ConsoleTextSizeY - 1;

	//adjust top-left-corner-postition
	if( ConsoleTopLeftX > ConsoleCursorX - ConsoleTextShowSizeX + 10 )
		ConsoleTopLeftX = ConsoleCursorX - ConsoleTextShowSizeX + 10;
	if( ConsoleTopLeftX < ConsoleCursorX - ConsoleTextShowSizeX + 1 )
		ConsoleTopLeftX = ConsoleCursorX - ConsoleTextShowSizeX + 1;
	if( ConsoleTopLeftY > ConsoleCursorY - 2 )
		ConsoleTopLeftY = ConsoleCursorY - 2;
	if( ConsoleTopLeftY < ConsoleCursorY - ConsoleTextShowSizeY + 1 )
		ConsoleTopLeftY = ConsoleCursorY - ConsoleTextShowSizeY + 1;

	//correct top-left-corner-position
	if( ConsoleTopLeftX < 0 )
		ConsoleTopLeftX = 0;
	if( ConsoleTopLeftX > ConsoleTextSizeX - ConsoleTextShowSizeX )
		ConsoleTopLeftX = ConsoleTextSizeX - ConsoleTextShowSizeX;
	if( ConsoleTopLeftY < 0 )
		ConsoleTopLeftY = 0;
	if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
		ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
}

//print a character to the console-buffer
static bool ConsolePut( char Chr )
{
	//check (and correct) position in console
	ConsoleCheckPos( );

	//no space in line
	if( ConsoleCursorX >= ConsoleTextSizeX - 2 )
		return false;

	//put character into text-buffer
	ConsoleText[ConsoleCursorY][ConsoleCursorX] = Chr;

	//go to next position in line
	ConsoleCursorX++;

	//check (and correct) position in console and adjust so that cursor is visible
	ConsoleCheckPosCursorVisible( );

	//console has changed
	ConsoleChanged = true;

	return true;
}

//remove a character from the console-buffer
static bool ConsoleBack( )
{
	//check (and correct) position in console
	ConsoleCheckPos( );

	//no character in line
	if( ConsoleCursorX <= 0 )
		return false;

	//go to previos position
	ConsoleCursorX--;

	//remove character from text-buffer
	ConsoleText[ConsoleCursorY][ConsoleCursorX] = 0;

	//check (and correct) position in console and adjust so that cursor is visible
	ConsoleCheckPosCursorVisible( );

	//console has changed
	ConsoleChanged = true;

	return true;
}

//go to next line in console-buffer
static void ConsoleNewline( )
{
	//local variables
	long X, Y;

	//check (and correct) position in console
	ConsoleCheckPos( );

	//log this line
	Log( "Console", ConsoleText[ConsoleCursorY] );

	//last line
	if( ConsoleCursorY == ConsoleTextSizeY - 1 )
	{
		//scroll up
		ConsoleCursorY--;
		for( Y = 0; Y < ConsoleTextSizeY - 1; Y++ )
		for( X = 0; X < ConsoleTextSizeX; X++ )
			ConsoleText[Y][X] = ConsoleText[Y + 1][X];
		for( X = 0; X < ConsoleTextSizeX; X++ )
			ConsoleText[ConsoleTextSizeY - 1][X] = 0;
	}

	//go to start of next line
	ConsoleCursorX = 0;
	ConsoleCursorY++;

	//check (and correct) position in console and adjust so that cursor is visible
	ConsoleCheckPosCursorVisible( );

	//console has changed
	ConsoleChanged = true;
}

//print an interpreted character to the console-buffer
static void ConsolePutc( char Chr )
{
	//backspace
	if( Chr == '\b' )
		ConsoleBack( );
	//newline
	if( Chr == '\n' )
		ConsoleNewline( );
	//normal character
	if( Chr >= ' ' )
		ConsolePut( Chr );
}

//print text to the console-buffer
static void ConsolePrint( char * Text )
{
	//print text
	for( ; *Text != 0; Text++ )
		ConsolePutc( *Text );
}

//insert a line in the console-buffer (just above the actual line)
static void ConsoleInsertLine( char * Text )
{
	//local variables
	long X, Y;

	//check (and correct) position in console
	ConsoleCheckPos( );

	//log this line
	Log( "Console", Text );

	//last line
	if( ConsoleCursorY == ConsoleTextSizeY - 1 )
	{
		//scroll up
		ConsoleCursorY--;
		for( Y = 0; Y < ConsoleTextSizeY - 1; Y++ )
		for( X = 0; X < ConsoleTextSizeX; X++ )
			ConsoleText[Y][X] = ConsoleText[Y + 1][X];
		for( X = 0; X < ConsoleTextSizeX; X++ )
			ConsoleText[ConsoleTextSizeY - 1][X] = 0;
	}

	//copy current line into next line
	for( X = 0; X < ConsoleTextSizeX; X++ )
		ConsoleText[ConsoleCursorY + 1][X] = ConsoleText[ConsoleCursorY][X];

	//this instruction prevents the microsoft c compiler in visual studio 6
	//to mix up some registers when compiling in release mode ?!?!?
	ConsoleText[ConsoleCursorY][0] += 0;

	//put text into current line
	for( X = 0; Text[X] != 0 && X < ConsoleTextSizeX; X++ )
		ConsoleText[ConsoleCursorY][X] = Text[X];
	//clear rest of line
	for( ; X < ConsoleTextSizeX; X++ )
		ConsoleText[ConsoleCursorY][X] = 0;

	//go to next line but stay in current column
	ConsoleCursorY++;

	//check (and correct) position in console and adjust so that cursor is visible
	ConsoleCheckPosCursorVisible( );

	//console has changed
	ConsoleChanged = true;
}

//refresh console
static void ConsoleRefresh( )
{
	//local variables
	long X, Y, FntX, FntY, TexX, TexY, ChrIndex;
	char Chr;
	unsigned char LineData;

	//check (and correct) position in console
	ConsoleCheckPos( );

	//put characters in console-buffer
	for( Y = 0; Y < ConsoleTextShowSizeY; Y++ )
	for( X = 0; X < ConsoleTextShowSizeX; X++ )
	{
		//character changed
		if( ConsoleTextShow[Y][X] != ConsoleText[Y + ConsoleTopLeftY][X + ConsoleTopLeftX] )
		{
			//get new character
			Chr = ConsoleText[Y + ConsoleTopLeftY][X + ConsoleTopLeftX];
			//put character into text-show-buffer
			ConsoleTextShow[Y][X] = Chr;
			//put character into console-buffer
			ChrIndex = (unsigned char)Chr;
			for( FntY = 0; FntY < FontSizeY; FntY++ )
			{
				LineData = FontData[ChrIndex][FntY];
				TexY = ConsoleInnerSizeY - (Y * FontSizeY + FntY) + ConsoleBorder;
				for( FntX = 0; FntX < FontSizeX; FntX++ )
				{
					TexX = X * FontSizeX + FntX + ConsoleBorder;
					if( (LineData & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
					{
						ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleTextB;
						ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleTextG;
						ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleTextR;
					}
					else
					{
						ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
						ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
						ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
					}
				}
			}
		}
	}

	//remove old cursor from console-buffer
	if( ConsoleCursorShowX >= 0 && ConsoleCursorShowX < ConsoleTextShowSizeX &&
		ConsoleCursorShowY >= 0 && ConsoleCursorShowY < ConsoleTextShowSizeY )
	{
		//get character
		ChrIndex = (unsigned char)ConsoleTextShow[ConsoleCursorShowY][ConsoleCursorShowX];
		//put character into console-buffer
		for( FntY = 0; FntY < FontSizeY; FntY++ )
		{
			LineData = FontData[ChrIndex][FntY];
			TexY = ConsoleInnerSizeY - (ConsoleCursorShowY * FontSizeY + FntY) + ConsoleBorder;
			for( FntX = 0; FntX < FontSizeX; FntX++ )
			{
				TexX = ConsoleCursorShowX * FontSizeX + FntX + ConsoleBorder;
				if( (LineData & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
				{
					ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleTextB;
					ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleTextG;
					ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleTextR;
				}
				else
				{
					ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
					ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
					ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
				}
			}
		}
	}

	//put cursor into console-buffer
	ConsoleCursorShowX = ConsoleCursorX - ConsoleTopLeftX;
	ConsoleCursorShowY = ConsoleCursorY - ConsoleTopLeftY;
	if( ConsoleCursorShowX >= 0 && ConsoleCursorShowX < ConsoleTextShowSizeX &&
		ConsoleCursorShowY >= 0 && ConsoleCursorShowY < ConsoleTextShowSizeY )
	{
		for( FntY = 0; FntY < FontSizeY; FntY++ )
		{
			LineData = CursorData[FntY];
			TexY = ConsoleInnerSizeY - (ConsoleCursorShowY * FontSizeY + FntY) + ConsoleBorder;
			for( FntX = 0; FntX < FontSizeX; FntX++ )
			{
				TexX = ConsoleCursorShowX * FontSizeX + FntX + ConsoleBorder;
				if( (LineData & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
				{
					ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleCursorB;
					ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleCursorG;
					ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleCursorR;
				}
				else
				{
					ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
					ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
					ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
				}
			}
		}
	}

	//load console-buffer into textrue
	LoadTexture( ConsoleData, ConsoleTexSizeX, ConsoleTexSizeY, ConsoleTexFirstNo );

	//console is refreshed
	ConsoleChanged = false;
}

//refresh console
static void GetConsoleLine( char * pLine, long LineSize )
{
	//local variables
	long I;

	//check (and correct) position in console
	ConsoleCheckPos( );

	//get number of characters in last line
	I = strlen( ConsoleText[ConsoleCursorY] );
	
	//do not exceed maximum-size
	if( I > LineSize - 1 )
		I = LineSize - 1;

	//copy last line
	memcpy( pLine, ConsoleText[ConsoleCursorY], I );
	pLine[I] = 0;
}

//load background from bmp-file
static bool LoadBackBmp( char * FileName )
{
	//local variables
	unsigned char * pData[BackTexCnt];
	long X, Y, I;

	//load bitmap in pieces
	if( !LoadImagePieces( FileName, BackTexSizeX, BackTexSizeY, BackTexCntX, BackTexCntY, pData, LoadBitmapFile ) )
		//error
		return false;

	//load bitmap-pieces into textures
	for( Y = 0; Y < BackTexCntY; Y++ )
	for( X = 0; X < BackTexCntX; X++ )
	{
		//calculate texture-number
		I = Y * BackTexCntX + X;
		//load texture
		LoadTexture( pData[I], BackTexSizeX, BackTexSizeY, BackTexFirstNo + I );
		//free data
		free( pData[I] );
	}

	//success
	return true;
}

//load background from jpg-file
static bool LoadBackJpg( char * FileName )
{
	//local variables
	unsigned char * pData[BackTexCnt];
	long X, Y, I;

	//load jpeg in pieces
	if( !LoadImagePieces( FileName, BackTexSizeX, BackTexSizeY, BackTexCntX, BackTexCntY, pData, LoadJpegFile ) )
		//error
		return false;

	//load bitmap-pieces into textures
	for( Y = 0; Y < BackTexCntY; Y++ )
	for( X = 0; X < BackTexCntX; X++ )
	{
		//calculate texture-number
		I = Y * BackTexCntX + X;
		//load texture
		LoadTexture( pData[I], BackTexSizeX, BackTexSizeY, BackTexFirstNo + I );
		//free data
		free( pData[I] );
	}

	//success
	return true;
}

//load texture from bmp-file
static bool LoadTexBmp( char * FileName, long TexNo, long TexSizeX, long TexSizeY )
{
	//local variables
	unsigned char * pData;

	//load image
	if( !LoadImageResize( FileName, TexSizeX, TexSizeY, &pData, LoadBitmapFile ) )
		//error
		return false;

	//load data into texture
	LoadTexture( pData, TexSizeX, TexSizeY, TexNo );

	//free data
	free( pData );

	//success
	return true;
}

//load texture from jpg-file
static bool LoadTexJpg( char * FileName, long TexNo, long TexSizeX, long TexSizeY  )
{
	//local variables
	unsigned char * pData;

	//load bitmap
	if( !LoadImageResize( FileName, TexSizeX, TexSizeY, &pData, LoadJpegFile ) )
		//error
		return false;

	//load data into texture
	LoadTexture( pData, TexSizeX, TexSizeY, TexNo );

	//free data
	free( pData );

	//success
	return true;
}

//load a script-file
static long LoadScript( char * FileName )
{
	//local variables
	FILE * File;
	char Line[1024];
	char * pChr;
	long I, L, Cnt;

	//script-buffer too large
	if( strlen( ScriptBuffer ) >= ScriptBufferSizeMax )
		return -1;
	
	//empty filename ===> fopen will raise assertion, so exit
	if( FileName == NULL || FileName[0] == 0 )
		return -1;

	//open file
	File = fopen( FileName, "rt" );
	if( File == NULL )
		return -1;

	//script-buffer not empty
	if( ScriptBuffer[0] != 0 )
	{
		//script-buffer does not end with newline
		L = strlen( ScriptBuffer );
		if( ScriptBuffer[L - 1] != '\n' )
		{
			//add a newline
			ScriptBuffer = (char*)realloc( ScriptBuffer, L + 2 );
			ScriptBuffer[L] = '\n';
			ScriptBuffer[L + 1] = 0;
		}
	}
	
	//read entire file line by line
	Cnt = 0;
	while( !feof( File ) )
	{
		//read a line
		Line[0] = 0;
		fgets( Line, sizeof( Line ), File );
		//replace newline with 0
		pChr = strchr( Line, '\n' );
		if( pChr != NULL )
			*pChr = 0;
		//remove illegal characters from line
		L = strlen( Line );
		for( I = 0; Line[I] != 0; I++ )
			if( (unsigned char)Line[I] < ' ' )
				memmove( Line + I, Line + I + 1, L - I );
		//when line not empty
		if( Line[0] != 0 )
		{
			//add line and newline to script-buffer
			L = strlen( ScriptBuffer );
			ScriptBuffer = (char*)realloc( ScriptBuffer, L + strlen( Line ) + 2 );
			strcpy( ScriptBuffer + L, Line );
			strcat( ScriptBuffer + L, "\n" );
			//count lines
			Cnt++;
		}
	}

	//close file
	fclose( File );

	//return number of lines
	return Cnt;
}

//process a command
//return true for "long" commands, false for "short" commands
static int ProcessCommand( char * Command )
{
	//local variables
	char * pChr, * pVal;
	long I, J, K;
	char Text[256];
	bool FuRes;

	//help
	if( strnicmp( Command, "help", 4 ) == 0 )
	{
		//display help
		ConsoleInsertLine( "  Exit, Quit: end program" );
		ConsoleInsertLine( "  Start: go to startpoint" );
		ConsoleInsertLine( "  Pos [x=<x>] [y=<y>] [z=<z>]: set or view position" );
		ConsoleInsertLine( "  Front [x=<x>] [y=<y>] [z=<z>]: set or view front-direction" );
		ConsoleInsertLine( "  CamptShow: show cam-points" );
		ConsoleInsertLine( "  CamptClr: clear cam-points" );
		ConsoleInsertLine( "  CamptSet <n> [px=<>] [py=<>] [pz=<>] [fx=<>] [fy=<>] [fz=<>]: set cam-point n" );
		ConsoleInsertLine( "  CamptIns <n> [px=<>] [py=<>] [pz=<>] [fx=<>] [fy=<>] [fz=<>]: insert cam-point n" );
		ConsoleInsertLine( "    p?: position of cam, f?: front direction of cam" );
		ConsoleInsertLine( "  CamptDel <n>: delete cam-point n" );
		ConsoleInsertLine( "  CamptSpeed [<speed>]: waypoint-move-speed in cam-points per second" );
		ConsoleInsertLine( "  CamptRefTime [Time|<seconds>]: reference time for cam-movement" );
		ConsoleInsertLine( "  CamMove [on|off]: enable / disable movement of camera" );
		ConsoleInsertLine( "  Console [on|off]: show / hide the console" );
		ConsoleInsertLine( "  BackGr [on|off]: enable / disable background" );
		ConsoleInsertLine( "  AntiAlias [on|off]: enable / disable antialiasing" );
		ConsoleInsertLine( "  BackLoadBmp [<filename>]: load background from bmp-file" );
		ConsoleInsertLine( "  BackLoadJpg [<filename>]: load background from jpg-file" );
		ConsoleInsertLine( "  BackUnload: unload background" );
		ConsoleInsertLine( "  PixTexLoadBmp [<filename>]: load pixel texture from bmp-file" );
		ConsoleInsertLine( "  PixTexLoadJpg [<filename>]: load pixel texture from jpg-file" );
		ConsoleInsertLine( "  PixTexUnload: unload pixel texture t" );
		ConsoleInsertLine( "  Time [<seconds>]: get / set time" );
		ConsoleInsertLine( "  ScriptLoad <filename>: load a script-file with commands" );
		ConsoleInsertLine( "  ScriptUnload: stop script execution and forget loaded scripts" );
		ConsoleInsertLine( "  ScriptStart: go on with script execution" );
		ConsoleInsertLine( "  ScriptStop: halt script execution" );
		ConsoleInsertLine( "  ScriptWaitAbs <seconds>: pause script execution until specified time reached" );
		ConsoleInsertLine( "  ScriptWaitRel <seconds>: pause script execution for specified number of seconds" );
		//long command
		return true;
	}

	//end program
	if( strnicmp( Command, "exit", 4 ) == 0 || strnicmp( Command, "quit", 4 ) == 0 )
	{
		PostQuitMessage( 0 );
		//long command
		return true;
	}

	//start-position
	if( strnicmp( Command, "start", 5 ) == 0 )
	{
		SetVector( &Pos, CenterX, CenterY, CenterZ );
		SetVector( &Front, 0.0, 1.0, 0.0 );
		//short command
		return false;
	}

	//go to position
	if( strnicmp( Command, "pos", 3 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 3;
		for( ; *pChr == ' '; pChr++ );
		//new x-value
		pVal = stristr( pChr, "x=" );
		if( pVal != NULL )
			Pos.X = DoubleMinMax( atof( pVal + 2 ), PosMinX, PosMaxX );
		//new y-value
		pVal = stristr( pChr, "y=" );
		if( pVal != NULL )
			Pos.Y = DoubleMinMax( atof( pVal + 2 ), PosMinY, PosMaxY );
		//new z-value
		pVal = stristr( pChr, "z=" );
		if( pVal != NULL )
			Pos.Z = DoubleMinMax( atof( pVal + 2 ), PosMinZ, PosMaxZ );
		//display position
		sprintf( Text, "  position is (%f, %f, %f)", Pos.X, Pos.Y, Pos.Z );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//set front-direction
	if( strnicmp( Command, "front", 5 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 5;
		for( ; *pChr == ' '; pChr++ );
		//new x-value
		pVal = stristr( pChr, "x=" );
		if( pVal != NULL )
			Front.X = DoubleMinMax( atof( pVal + 2 ), -1.0, 1.0 );
		//new y-value
		pVal = stristr( pChr, "y=" );
		if( pVal != NULL )
			Front.Y = DoubleMinMax( atof( pVal + 2 ), -1.0, 1.0 );
		//new z-value
		pVal = stristr( pChr, "z=" );
		if( pVal != NULL )
			Front.Z = DoubleMinMax( atof( pVal + 2 ), -1.0, 1.0 );
		//repair position and front direction
		RepairPosFront( );
		//display front-direction
		sprintf( Text, "  front-direction is (%f, %f, %f)", Front.X, Front.Y, Front.Z );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//show cam-points
	if( strnicmp( Command, "camptshow", 9 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 9;
		for( ; *pChr == ' '; pChr++ );
		//display cam-points
		sprintf( Text, "  camera-points: %d", CamptsCnt );
		ConsoleInsertLine( Text );
		for( J = 0; J < CamptsCnt; J++ )
		{
			sprintf( Text, "  %3d: pos (%f, %f, %f) front (%f, %f, %f)", J,
				pCampts[J].Pos.X, pCampts[J].Pos.Y, pCampts[J].Pos.Z,
				pCampts[J].Front.X, pCampts[J].Front.Y, pCampts[J].Front.Z );
			ConsoleInsertLine( Text );
		}
		//short command
		return false;
	}

	//clear cam-points
	if( strnicmp( Command, "camptclr", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//clear cam-points
		CamptsCnt = 0;
		pCampts = (tCampt *)realloc( pCampts, 0 );
		//display message
		ConsoleInsertLine( "  camera-points cleared" );
		//short command
		return false;
	}

	//set cam-point n
	if( strnicmp( Command, "camptset", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get cam-point number
		J = atoi( pChr );
		//jump over number and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//error
		if( J < 0 || J > CamptsCnt - 1 )
		{
			sprintf( Text, "  camera-point %d does not exist", J );
			ConsoleInsertLine( Text );
		}
		//set cam-point
		else
		{
			//new x-value for pos
			pVal = stristr( pChr, "px=" );
			if( pVal != NULL )
				pCampts[J].Pos.X = DoubleMinMax( atof( pVal + 3 ), PosMinX, PosMaxX );
			//new y-value for pos
			pVal = stristr( pChr, "py=" );
			if( pVal != NULL )
				pCampts[J].Pos.Y = DoubleMinMax( atof( pVal + 3 ), PosMinY, PosMaxY );
			//new z-value for pos
			pVal = stristr( pChr, "pz=" );
			if( pVal != NULL )
				pCampts[J].Pos.Z = DoubleMinMax( atof( pVal + 3 ), PosMinZ, PosMaxZ );
			//new x-value for front
			pVal = stristr( pChr, "fx=" );
			if( pVal != NULL )
				pCampts[J].Front.X = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
			//new y-value for front
			pVal = stristr( pChr, "fy=" );
			if( pVal != NULL )
				pCampts[J].Front.Y = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
			//new z-value for front
			pVal = stristr( pChr, "fz=" );
			if( pVal != NULL )
				pCampts[J].Front.Z = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
			//repair position and front direction
			RepairPosFront( );
			//display message
			sprintf( Text, "  camera-point %d set to pos (%f, %f, %f) front (%f, %f, %f)", J,
				pCampts[J].Pos.X, pCampts[J].Pos.Y, pCampts[J].Pos.Z,
				pCampts[J].Front.X, pCampts[J].Front.Y, pCampts[J].Front.Z );
			ConsoleInsertLine( Text );
		}
		//short command
		return false;
	}

	//insert cam-point n
	if( strnicmp( Command, "camptins", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get waypoint number
		J = atoi( pChr );
		if( J < 0 )
			J = 0;
		if( J > CamptsCnt )
			J = CamptsCnt;
		//jump over number and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//insert cam-point
		CamptsCnt++;
		pCampts = (tCampt *)realloc( pCampts, CamptsCnt * sizeof( tCampt ) );
		for( K = CamptsCnt - 1; K > J; K-- )
		{
			Copy( &pCampts[K].Pos, pCampts[K-1].Pos );
			Copy( &pCampts[K].Front, pCampts[K-1].Front );
		}
		Copy( &pCampts[J].Pos, Pos );
		Copy( &pCampts[J].Front, Front );
		//new x-value for pos
		pVal = stristr( pChr, "px=" );
		if( pVal != NULL )
			pCampts[J].Pos.X = DoubleMinMax( atof( pVal + 3 ), PosMinX, PosMaxX );
		//new y-value for pos
		pVal = stristr( pChr, "py=" );
		if( pVal != NULL )
			pCampts[J].Pos.Y = DoubleMinMax( atof( pVal + 3 ), PosMinY, PosMaxY );
		//new z-value for pos
		pVal = stristr( pChr, "pz=" );
		if( pVal != NULL )
			pCampts[J].Pos.Z = DoubleMinMax( atof( pVal + 3 ), PosMinZ, PosMaxZ );
		//new x-value for front
		pVal = stristr( pChr, "fx=" );
		if( pVal != NULL )
			pCampts[J].Front.X = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
		//new y-value for front
		pVal = stristr( pChr, "fy=" );
		if( pVal != NULL )
			pCampts[J].Front.Y = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
		//new z-value for front
		pVal = stristr( pChr, "fz=" );
		if( pVal != NULL )
			pCampts[J].Front.Z = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
		//repair position and front direction
		RepairPosFront( );
		//display message
		sprintf( Text, "  camera-point %d inserted and set to pos (%f, %f, %f) front (%f, %f, %f)", J,
			pCampts[J].Pos.X, pCampts[J].Pos.Y, pCampts[J].Pos.Z,
			pCampts[J].Front.X, pCampts[J].Front.Y, pCampts[J].Front.Z );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//delete cam-point n
	if( strnicmp( Command, "camptdel", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get cam-point number
		J = atoi( pChr );
		//jump over number and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//error
		if( J < 0 || J > CamptsCnt - 1 )
		{
			sprintf( Text, "  camera-point %d does not exist", J );
			ConsoleInsertLine( Text );
		}
		//delete cam-point
		else
		{
			CamptsCnt--;
			for( K = J; K < CamptsCnt; K++ )
			{
				Copy( &pCampts[K].Pos, pCampts[K+1].Pos );
				Copy( &pCampts[K].Front, pCampts[K+1].Front );
			}
			pCampts = (tCampt *)realloc( pCampts, CamptsCnt * sizeof( tCampt ) );
			//display message
			sprintf( Text, "  camera-point %d deleted", J );
			ConsoleInsertLine( Text );
		}
		//short command
		return false;
	}

	//cam-point-speed
	if( strnicmp( Command, "camptspeed", 10 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 10;
		for( ; *pChr == ' '; pChr++ );
		//new value
		if( *pChr != 0 )
			CamptSpeed = DoubleMinMax( atof( pChr ), -CamptSpeedRange, CamptSpeedRange );
		//display speed
		sprintf( Text, "  camera-point-move-speed is now %f waypoints per second", CamptSpeed );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//cam-point-reference-time
	if( strnicmp( Command, "camptreftime", 12 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 12;
		for( ; *pChr == ' '; pChr++ );
		//new value
		if( *pChr != 0 )
		{
			//current time
			if( strnicmp( pChr, "time", 4 ) == 0 )
				CamptRefTime = TimeGet( );
			//value in seconds
			else
				CamptRefTime = DoubleMinMax( atof( pChr ), -TimeRange, TimeRange );
		}
		//display message
		sprintf( Text, "  reference time for camera-point-movement is %f", CamptRefTime );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//camera moving or not
	if( strnicmp( Command, "cammove", 7 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 7;
		for( ; *pChr == ' '; pChr++ );
		//toggle
		CamMoving = !CamMoving;
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			CamMoving = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			CamMoving = false;
		//display state
		if( CamMoving )
			ConsoleInsertLine( "  camera is now moving" );
		else
			ConsoleInsertLine( "  camera is now stopped" );
		//short command
		return false;
	}

	//console
	if( strnicmp( Command, "console", 7 ) == 0 )
	{
		//toggle console
		Console = !Console;
		//jump over spaces
		pChr = Command + 7;
		for( ; *pChr == ' '; pChr++ );
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			Console = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			Console = false;
		//display state
		if( Console )
			ConsoleInsertLine( "  console is now turned on" );
		else
			ConsoleInsertLine( "  console is now turned off" );
		//short command
		return false;
	}

	//background
	if( strnicmp( Command, "backgr", 6 ) == 0 )
	{
		//toggle background
		Back = !Back;
		//jump over spaces
		pChr = Command + 6;
		for( ; *pChr == ' '; pChr++ );
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			Back = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			Back = false;
		//display state
		if( Back )
			ConsoleInsertLine( "  background is now turned on" );
		else
			ConsoleInsertLine( "  background is now turned off" );
		//short command
		return false;
	}

	//antialiasing
	if( strnicmp( Command, "antialias", 9 ) == 0 )
	{
		//toggle antialiasing
		AntiAlias = !AntiAlias;
		//jump over spaces
		pChr = Command + 9;
		for( ; *pChr == ' '; pChr++ );
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			AntiAlias = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			AntiAlias = false;
		//display state
		if( AntiAlias )
			ConsoleInsertLine( "  antialiasing is now turned on" );
		else
			ConsoleInsertLine( "  antialiasing is now turned off" );
		//short command
		return false;
	}

	//load background from bitmap
	if( strnicmp( Command, "backloadbmp", 11 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 11;
		for( ; *pChr == ' '; pChr++ );
		//load new background
		FuRes = LoadBackBmp( pChr );
		//display message
		if( FuRes )
			ConsoleInsertLine( "  background loaded" );
		else
			ConsoleInsertLine( "  background not loaded" );
		//long command
		return true;
	}

	//load background from jpeg
	if( strnicmp( Command, "backloadjpg", 11 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 11;
		for( ; *pChr == ' '; pChr++ );
		//load new background
		FuRes = LoadBackJpg( pChr );
		//display message
		if( FuRes )
			ConsoleInsertLine( "  background loaded" );
		else
			ConsoleInsertLine( "  background not loaded" );
		//long command
		return true;
	}

	//unload background
	if( strnicmp( Command, "backunload", 10 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 10;
		for( ; *pChr == ' '; pChr++ );
		//unload background-textures
		for( I = BackTexFirstNo; I < BackTexFirstNo + BackTexCnt; I++ )
			UnloadTexture( I );
		//display message
		ConsoleInsertLine( "  background unloaded" );
		//long command
		return true;
	}

	//load pixel texture from bitmap
	if( strnicmp( Command, "pixtexloadbmp", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//load new texture
		FuRes = LoadTexBmp( pChr, PixelTexFirstNo, PixTexSizeX, PixTexSizeY );
		//display message
		if( FuRes )
			sprintf( Text, "  pixel texture loaded" );
		else
			sprintf( Text, "  pixel texture not loaded" );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}
	
	//load pixel texture from jpeg
	if( strnicmp( Command, "pixtexloadjpg", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//load new texture
		FuRes = LoadTexJpg( pChr, PixelTexFirstNo, PixTexSizeX, PixTexSizeY );
		//display message
		if( FuRes )
			sprintf( Text, "  pixel texture loaded" );
		else
			sprintf( Text, "  pixel texture not loaded" );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//unload texture
	if( strnicmp( Command, "texunload", 9 ) == 0 )
	{
		//unload texture
		UnloadTexture( PixelTexFirstNo );
		//display message
		sprintf( Text, "  pixel texture unloaded" );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//set time
	if( strnicmp( Command, "time", 4 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 4;
		for( ; *pChr == ' '; pChr++ );
		//new value
		if( *pChr != 0 )
			TimeSet( DoubleMinMax( atof( pChr ), -TimeRange, TimeRange ) );
		//display message
		sprintf( Text, "  current time is %f", TimeGet( ) );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//load a script-file
	if( strnicmp( Command, "scriptload", 10 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 10;
		for( ; *pChr == ' '; pChr++ );
		//load script
		I = LoadScript( pChr );
		//display number of lines loaded
		if( I < 0 )
			sprintf( Text, "  script could not be loaded" );
		else
			sprintf( Text, "  script loaded (%d lines)", I );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//unload script
	if( strnicmp( Command, "scriptunload", 12 ) == 0 )
	{
		//empty script-buffer
		ScriptBuffer = (char*)realloc( ScriptBuffer, 1 );
		ScriptBuffer[0] = 0;
		//display message
		ConsoleInsertLine( "  script unloaded" );
		//long command
		return true;
	}

	//go on with script execution
	if( strnicmp( Command, "scriptstart", 11 ) == 0 )
	{
		//set script-mode to running
		ScriptMode = 'R';
		//display message
		ConsoleInsertLine( "  script execution goes on" );
		//long command
		return true;
	}

	//halt script execution
	if( strnicmp( Command, "scriptstop", 10 ) == 0 )
	{
		//set script-mode to stopped
		ScriptMode = 'S';
		//display message
		ConsoleInsertLine( "  script execution halted" );
		//long command
		return true;
	}

	//pause script absolute
	if( strnicmp( Command, "scriptwaitabs", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//get value
		if( *pChr != 0 )
			ScriptWaitTime = DoubleMinMax( atof( pChr ), -TimeRange, TimeRange );
		//set script-mode to waiting
		ScriptMode = 'W';
		//display message
		sprintf( Text, "  current time is %f", TimeGet( ) );
		ConsoleInsertLine( Text );
		sprintf( Text, "  script execution halted till time %f", ScriptWaitTime );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//pause script relative
	if( strnicmp( Command, "scriptwaitrel", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//get value
		if( *pChr != 0 )
			ScriptWaitTime = DoubleMinMax( atof( pChr ) + TimeGet( ), 0.0, TimeRange );
		//set script-mode to waiting
		ScriptMode = 'W';
		//display message
		sprintf( Text, "  current time is %f", TimeGet( ) );
		ConsoleInsertLine( Text );
		sprintf( Text, "  script execution halted till time %f", ScriptWaitTime );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//unknown command
	ConsoleInsertLine( "  unknown command (type \"help\" for list of commands)" );
	//long command
	return true;
}

//process a command from the script-buffer
//return true for "long" commands, false for "short" commands
static int ProcessScriptBufferCommand( )
{
	//local variables
	char * pChr, * pNl, * pCommand;
	char ConsoleLine[ConsoleTextSizeX];
	int LongCommand;

	//buffer empty?
	if( ScriptBuffer[0] == 0 )
		//long command
		return true;

	//jump over spaces and newlines
	for( pChr = ScriptBuffer; *pChr == ' ' || *pChr == '\n'; pChr++ );

	//end of buffer?
	if( *pChr == 0 )
	{
		//empty buffer
		ScriptBuffer[0] = 0;
		ScriptBuffer = (char *)realloc( ScriptBuffer, 1 );
		//long command
		return true;
	}

	//search for next newline
	pNl = strchr( pChr, '\n' );
	if( pNl == NULL ) //not found
	{
		//put command in extra-buffer
		pCommand = strdup( pChr );
		//empty script-buffer
		ScriptBuffer = (char *)realloc( ScriptBuffer, 1 );
		ScriptBuffer[0] = 0;
	}
	else //found
	{
		//put command in extra-buffer
		*pNl = 0; //terminate string before newline
		pCommand = strdup( pChr );
		//remove command from buffer and resize buffer
		pChr = pNl + 1;
		memmove( ScriptBuffer, pChr, strlen( pChr ) + 1 );
		ScriptBuffer = (char *)realloc( ScriptBuffer, strlen( ScriptBuffer ) + 1 );
	}

	//display command
	strcpy( ConsoleLine, "[script] " );
	strncat( ConsoleLine, pCommand, sizeof( ConsoleLine ) - strlen( ConsoleLine ) - 1 );
	ConsoleInsertLine( ConsoleLine );
	//process command (ATTENTION: could append something to ScriptBuffer)
	LongCommand = ProcessCommand( pCommand );
	//free command-string
	free( pCommand );

	//long command?
	return LongCommand;
}

//process a few command from the script-buffer
static void ProcessScriptBufferCommands( )
{
	//local variables
	long I;

	//script-mode is waiting
	if( ScriptMode == 'W' )
	{
		//time reached => switch to running
		if( TimeGet( ) >= ScriptWaitTime )
			ScriptMode = 'R';
	}

	//script-mode is running
	if( ScriptMode == 'R' )
	{
		//process max MaxScriptCommandsAtSameTime commands
		for( I = 0; I < MaxScriptCommandsAtSameTime; I++ )
			//process command
			if( ProcessScriptBufferCommand( ) )
				//long command => no more commands
				break;
	}
}

//mouse-steering
static void Mouse( )
{
	//local variables
	POINT MousePos;
	long MouseHor, MouseVer;
	bool MouseLeft, MouseRight;
	tVector Dir, Top, Left;

	//get direction over ground
	Dir = Front;
	Dir.Z = 0;
	Normalize( &Dir );
	//get top and left vectors
	SetVector( &Top, 0.0, 0.0, 1.0 );
	Left = Dir;
	Rotate( &Left, Top, -0.5 * Pi );

	//get mouse-movement
	GetCursorPos( &MousePos );
	MouseHor = MousePos.x - ScreenX / 2;
	MouseVer = MousePos.y - ScreenY / 2;
	//get mouse-buttons
	MouseLeft = (GetAsyncKeyState( VK_LBUTTON ) & 0x8000) != 0;
	MouseRight = (GetAsyncKeyState( VK_RBUTTON ) & 0x8000) != 0;

	if( MouseLeft )
	{
		if( !MouseRight )
		{
			//move forwards / backwards
			LinearCombine( &Pos, 1.0f, Pos, - MouseVer * MouseMoveStep, Dir );
			//move left / right
			LinearCombine( &Pos, 1.0f, Pos, - MouseHor * MouseMoveStep, Left );
		}
	}
	else
	{
		if( MouseRight )
		{
			//move up / down
			LinearCombine( &Pos, 1.0f, Pos, MouseVer * MouseMoveStep, Top );
		}
		else
		{
			//turn left / right
			Rotate( &Front, Top, MouseHor * MouseTurnStep );
			//look up / down
			Rotate( &Front, Left, - MouseVer * InvertMouse * MouseTurnStep );
		}
	}

	//repair position and front direction
	RepairPosFront( );

	//center mousepointer
	SetCursorPos( ScreenX / 2, ScreenY / 2 );
}

//window procedure
LRESULT CALLBACK WindowProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	//local variables
	char Line[256];
	tVector Dir, Top, Left;

	switch( uMsg )
	{
	//key was pressed
	case WM_KEYDOWN:
		//get direction over ground
		Dir = Front;
		Dir.Z = 0;
		Normalize( &Dir );
		//get top and left vectors
		SetVector( &Top, 0.0, 0.0, 1.0 );
		Left = Dir;
		Rotate( &Left, Top, -0.5 * Pi );

		//distinguish between keys
		switch( wParam )
		{
		//escape
		case VK_ESCAPE:
			//toggle console
			Console = !Console;
			//display message
			if( Console )
				ConsoleInsertLine( "[escape] => console on" );
			else
				ConsoleInsertLine( "[escape] => console off" );
			break;
		//enter
		case VK_RETURN:
			if( ! Console ) //console off
			{
				//go to start position
				SetVector( &Pos, CenterX, CenterY, CenterZ );
				SetVector( &Front, 0.0, 1.0, 0.0 );
			}
			break;
		//space
		case VK_SPACE:	
			if( ! Console ) //console off
			{
				//restore horizontal view
				Front.Z = 0.0;
			}
			break;
		//left arrow
		case VK_LEFT:
			if( Console ) //console on
			{
				//move console left
				ConsoleTopLeftX -= ConsoleLeftRightStep;
				if( ConsoleTopLeftX < 0 )
					ConsoleTopLeftX = 0;
				ConsoleChanged = true;
			}
			else //console off
			{
				//turn left
				Rotate( &Front, Top, - TurnStep );
			}
			break;
		//right arrow
		case VK_RIGHT:
			if( Console ) //console on
			{
				//move console right
				ConsoleTopLeftX += ConsoleLeftRightStep;
				if( ConsoleTopLeftX > ConsoleTextSizeX - ConsoleTextShowSizeX )
					ConsoleTopLeftX = ConsoleTextSizeX - ConsoleTextShowSizeX;
				ConsoleChanged = true;
			}
			else //console off
			{
				//turn right
				Rotate( &Front, Top, TurnStep );
			}
			break;
		//up arrow
		case VK_UP:
			if( Console ) //console on
			{
				//move console up
				ConsoleTopLeftY -= ConsoleUpDownStep;
				if( ConsoleTopLeftY < 0 )
					ConsoleTopLeftY = 0;
				ConsoleChanged = true;
			}
			else //console off
			{
				//move forwards
				LinearCombine( &Pos, 1.0f, Pos, MoveStep, Dir );
			}
			break;
		//down arrow
		case VK_DOWN:
			if( Console ) //console on
			{
				//move console down
				ConsoleTopLeftY += ConsoleUpDownStep;
				if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
					ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
				ConsoleChanged = true;
			}
			else //console off
			{
				//move backwards
				LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Dir );
			}
			break;
		//home
		case VK_HOME:
			if( ! Console ) //console off
			{
				//look up
				Rotate( &Front, Left, TurnStep );
			}
			break;
		//end
		case VK_END:
			if( ! Console ) //console off
			{
				//look down
				Rotate( &Front, Left, - TurnStep );
			}
			break;
		//page up
		case 0x21: //VK_PAGE_UP
			if( Console ) //console on
			{
				//move console up a page
				ConsoleTopLeftY -= ConsolePageStep;
				if( ConsoleTopLeftY < 0 )
					ConsoleTopLeftY = 0;
				ConsoleChanged = true;
			}
			else //console off
			{
				//move up
				LinearCombine( &Pos, 1.0f, Pos, MoveStep, Top );
			}
			break;
		//page down
		case 0x22: //VK_PAGE_DOWN
			if( Console ) //console on
			{
				//move console down a page
				ConsoleTopLeftY += ConsolePageStep;
				if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
					ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
				ConsoleChanged = true;
			}
			else //console off
			{
				//move down
				LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Top );
			}
			break;
		//insert
		case VK_INSERT:
			if( !Console ) //console off
			{
				//move left
				LinearCombine( &Pos, 1.0f, Pos, MoveStep, Left );
			}
			break;
		//delete
		case VK_DELETE:
			if( !Console ) //console off
			{
				//move right
				LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Left );
			}
			break;
		}

		//repair position and front direction
		RepairPosFront( );
		break;

	//character was typed
	case WM_CHAR:
		//console is inactive ===> do nothing
		if( !Console )
			break;
		//normal characters
		if( (unsigned char)wParam >= ' ' )
			ConsolePut( (char)wParam );
		//enter
		if( wParam == '\r' )
		{
			//get last line of console
			GetConsoleLine( Line, sizeof( Line ) );
			//display enter
			ConsoleNewline( );
			//process Command
			ProcessCommand( Line );
		}
		//backspace
		if( wParam == '\b' )
			ConsoleBack( );
		break;

	//window-close-click
	case WM_CLOSE:
		//end
		PostQuitMessage( 0 );
		break;
	}
	
	return DefWindowProc( hWnd, uMsg, wParam, lParam );
}

//main program
int WINAPI WinMain( HINSTANCE AppHndl, HINSTANCE UnusedHndl, LPSTR Parameters, int ShowMode )
{
	//local variables
	long I, J, X, Y, Z;
	DEVMODE DevMode;
	WNDCLASS WindowClass;
	int PixelFormat;
	PIXELFORMATDESCRIPTOR PixelFormatDesc = 
	{
		sizeof( PIXELFORMATDESCRIPTOR ), //size of structure
		1, //version
		PFD_DRAW_TO_WINDOW | //supports Window
		PFD_SUPPORT_OPENGL | //supports OpenGl
		PFD_DOUBLEBUFFER, //doublebuffered
		PFD_TYPE_RGBA, //RGBA-type
		16, //16 bit color-depth
		0, 0, 0, 0, 0, 0, //ignore color-bits
		0, //no alpha-buffer
		0, //no shift-bit
		0, //no accumulation-buffer
		0, 0, 0, 0, //ignore accumulation-bits
		16, //16 bit Z-buffer
		0, //no stencil-buffer
		0, //no auxiluary buffer
		PFD_MAIN_PLANE, //main plane
		0, //reserved
		0, 0, 0 //ignore layer-masks
	};
	char Text[256], * pChr;
	MSG Message;

	//global initialisations
	AppHandle = AppHndl;
	srand( GetTickCount( ) );

	//open logfile
	if( stristr( Parameters, "Log" ) != NULL )
		LogFile = fopen( "TroiaPreview.log", "wt" );
	else
		LogFile = NULL;

	//resolution - autodetect
	ScreenX = GetSystemMetrics( SM_CXSCREEN );
	ScreenY = GetSystemMetrics( SM_CYSCREEN );
	//resolution - parameters
	if( (pChr = stristr( Parameters, "Res" )) != NULL )
	{
		I = 640;
		J = 480;
		if( sscanf( pChr+3, "%dx%d", &X, &Y ) == 2 )
		{
			ScreenX = LongMinMax( X, 640, 6400 );
			ScreenY = LongMinMax( Y, 400, 4000 );
		}
	}

	//mouse-mode on ?
	MouseModeOn = stristr( Parameters, "NoMouse" ) == NULL;
	//invert mouse ?
	if( stristr( Parameters, "InvertMouse" ) != NULL )
		InvertMouse = -1;

	//full-screen-mode on
	FullScreen = stristr( Parameters, "NoFullScreen" ) == NULL;
	if( FullScreen )
	{
		DevMode.dmSize = sizeof( DevMode );
		DevMode.dmBitsPerPel = 16;
		DevMode.dmPelsWidth = ScreenX;
		DevMode.dmPelsHeight = ScreenY;
		DevMode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		ChangeDisplaySettings( &DevMode, CDS_FULLSCREEN );
	}

	//register window-class
	WindowClass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	WindowClass.lpfnWndProc = &WindowProc; 
	WindowClass.cbClsExtra = 0;
	WindowClass.cbWndExtra = 0;
	WindowClass.hInstance = AppHandle;
	WindowClass.hIcon = NULL;
	WindowClass.hCursor = NULL;
	WindowClass.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
	WindowClass.lpszMenuName = NULL;
	WindowClass.lpszClassName = "TroiaPreview";
	RegisterClass( &WindowClass );

	//create window
	WindowHandle = CreateWindow( "TroiaPreview", "TroiaPreview", WS_POPUP | WS_VISIBLE, 0, 0, ScreenX, ScreenY, NULL, NULL, AppHandle, NULL );
	while( ShowCursor( false ) >= 0 );
	SetCursor( NULL );

	//initialize OpenGl
	WindowDeviceHandle = GetDC( WindowHandle );
	PixelFormat = ChoosePixelFormat( WindowDeviceHandle, &PixelFormatDesc );
	SetPixelFormat( WindowDeviceHandle, PixelFormat, &PixelFormatDesc );
	DescribePixelFormat( WindowDeviceHandle, PixelFormat, sizeof( PixelFormatDesc ), &PixelFormatDesc );
	OpenGlContextHandle = wglCreateContext( WindowDeviceHandle );
	wglMakeCurrent( WindowDeviceHandle, OpenGlContextHandle );

	//OpenGl settings
	glDepthFunc( GL_LEQUAL );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_AUTO_NORMAL );
	glEnable( GL_NORMALIZE );
	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
	glViewport( 0, 0, ScreenX, ScreenY );
	glLineWidth( 1.5 ); //set line width
	glPointSize( 2.0 ); //set point size

	//initialize cam-points
	CamptsCnt = 0; //no cam-points
	pCampts = (tCampt *)malloc( 0 );
	CamptSpeed = 0.0; //cam-point-speed
	CamptRefTime = 0.0; //reference-time for cam-point-movement
	CamMoving = false; //if cam is not moving

	//start-position
	SetVector( &Pos, CenterX, CenterY, CenterZ );
	SetVector( &Front, 0.0, 1.0, 0.0 );

	//initialise console
	ConsoleInit( );
	ConsoleInsertLine( "OpenGL information" );
	strcpy( Text, "  GL_VENDOR: " );
	strncat( Text, (char *)glGetString( GL_VENDOR ), sizeof( Text ) - strlen( Text ) - 1 );
	ConsoleInsertLine( Text );
	strcpy( Text, "  GL_RENDERER: " );
	strncat( Text, (char *)glGetString( GL_RENDERER ), sizeof( Text ) - strlen( Text ) - 1 );
	ConsoleInsertLine( Text );
	strcpy( Text, "  GL_VERSION: " );
	strncat( Text, (char *)glGetString( GL_VERSION ), sizeof( Text ) - strlen( Text ) - 1 );
	ConsoleInsertLine( Text );
	ConsoleInsertLine( "TroiaPreview" );
	ConsoleInsertLine( "  version 0.1 date 2004-09-01" );
	ConsoleInsertLine( "  Copyright (C) 2003-2004 1stein <1stein@schuermans.info>" );
	ConsoleInsertLine( "  Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html" );
	ConsoleInsertLine( "  some important keys:" );
	ConsoleInsertLine( "    alt-f4: exit program" );
	ConsoleInsertLine( "    escape: toggle console" );
	ConsoleInsertLine( "  when console is off:" );
	ConsoleInsertLine( "    arrow-keys: move around" );
	ConsoleInsertLine( "  when console is on:" );
	ConsoleInsertLine( "    arrow-keys: scroll console" );
	ConsoleInsertLine( "  type \"help\" for commands" );
	//initialize script-buffer
	ScriptBuffer = strdup( "scriptload TroiaPreview.ini" );

	//initialize current frame
	for( Z = 0; Z < PixelY; Z++ )
	{
		for( X = 0; X < PixelX; X++ )
		{
			Wall1[Z][X] = (unsigned char)rand( );
			Wall3[Z][X] = (unsigned char)rand( );
		}
		for( Y = 0; Y < PixelY; Y++ )
		{
			Wall2[Z][Y] = (unsigned char)rand( );
			Wall4[Z][Y] = (unsigned char)rand( );
		}
	}
	for( Y = 0; Y < PixelY; Y++ )
	{
		for( X = 0; X < PixelX; X++ )
		{
			Ceiling[Y][X] = (unsigned char)rand( );
			Floor[Y][X] = (unsigned char)rand( );
		}
	}

	//center mousepointer
	if( MouseModeOn )
		SetCursorPos( ScreenX / 2, ScreenY / 2 );

	//set time to 0
	TimeSet( 0.0 );

	//message-loop
	Message.message = 0;

	while( Message.message != WM_QUIT )
	{
		//process all messages
		while( PeekMessage( &Message, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &Message );
			DispatchMessage( &Message );
			if( Message.message == WM_QUIT )
				break;
		}
		//mouse-steering
		if( MouseModeOn )
			Mouse( );
		//refresh console
		if( Console && ConsoleChanged )
			ConsoleRefresh( );
		//show new picture
		Show( );
		//process a few script-commands
		ProcessScriptBufferCommands( );
		//move camera
		MoveCam( );
	}

	//free script-buffer
	free( ScriptBuffer );
	
	//unload cam-points
	free( pCampts );

	//cleanup OpenGl
	wglMakeCurrent( NULL, NULL );
	wglDeleteContext( OpenGlContextHandle );
	ReleaseDC( WindowHandle, WindowDeviceHandle );

	//destroy window and windows-class
	ShowCursor( true );
	DestroyWindow( WindowHandle );
	UnregisterClass( "TroiaPreview", AppHandle );

	//full-screen-mode-off
	if( FullScreen )
		ChangeDisplaySettings( NULL, 0 );

	//close logfile
	if( LogFile != NULL )
		fclose( LogFile );

	//end
    return 0;
}
